"""Extended phase graph (EPG) signal models."""

# Based on https://github.com/fzimmermann89/epgtorch/
# Copyright (c) 2022 Felix Zimmermann, MIT License

from abc import ABC, abstractmethod
from collections.abc import Sequence

import torch

from mrpro.data.Dataclass import Dataclass
from mrpro.utils.reshape import unsqueeze_tensors_right
from mrpro.utils.TensorAttributeMixin import TensorAttributeMixin


def move_tensors_to_gpu_if_any_on_gpu(*tensors: torch.Tensor) -> tuple[torch.Tensor, ...]:
    """Move tensors to the GPU if any of them is on the GPU.

    Parameters
    ----------
    tensors
        Tensors to move to the GPU if any of them is on the GPU. If none of the tensors is on the GPU, they are
        returned unchanged.

    Returns
    -------
        List of tensors moved to the GPU if any of them is on the GPU, otherwise the original tensors.
    """
    for t in tensors:
        if t.is_cuda:
            device = t.device
            return tuple(tensor.to(device) for tensor in tensors)
    return tensors


class Parameters(Dataclass):
    """Tissue parameters for EPG simulation."""

    m0: torch.Tensor
    """Steay state magnetization (complex)"""

    t1: torch.Tensor
    """T1 relaxation time [s]"""

    t2: torch.Tensor
    """T2 relaxation time [s]"""

    relative_b1: torch.Tensor | None = None
    """Relative B1 scaling factor (complex)"""

    t1_rho: torch.Tensor | None = None
    """T1 rho relaxation time [s]"""

    @property
    def device(self) -> torch.device:
        """Device of the parameters."""
        device = super().device
        assert device is not None  # mypy # noqa: S101
        return device

    @property
    def shape(self) -> torch.Size:
        """Broadcasted shape of the parameters."""
        shape = torch.broadcast_shapes(self.t1.shape, self.t2.shape, self.m0.shape)
        if self.relative_b1 is not None:
            shape = torch.broadcast_shapes(shape, self.relative_b1.shape)
        return shape

    @property
    def dtype(self) -> torch.dtype:
        """Promoted data type of the parameters."""
        dtype = torch.promote_types(self.t1.dtype, self.t2.dtype)
        dtype = torch.promote_types(dtype, self.m0.dtype)
        if self.relative_b1 is not None:
            dtype = torch.promote_types(dtype, self.relative_b1.dtype)
        return dtype

    @property
    def ndim(self) -> int:
        """Number of dimensions of the parameters."""
        ndim = max(self.t1.ndim, self.t2.ndim, self.m0.ndim)
        if self.relative_b1 is not None:
            ndim = max(ndim, self.relative_b1.ndim)
        return ndim


def rf_matrix(
    flip_angle: torch.Tensor,
    phase: torch.Tensor,
    relative_b1: torch.Tensor | None = None,
) -> torch.Tensor:
    """Initialize the rotation matrix describing the RF pulse.

    Parameters
    ----------
    flip_angle
        Flip angle of the RF pulse in rad
    phase
        Phase of the RF pulse in rad
    relative_b1
        Scaling of flip angle due to B1 inhomogeneities

    Returns
    -------
        Matrix describing the mixing of the EPG configuration states due to an RF pulse.
    """
    if relative_b1 is not None:
        # relative_b1 and flip_angle need to be correctly broadcasted at this point
        flip_angle = flip_angle * relative_b1.abs()
        phase = phase + relative_b1.angle()
    cosa = torch.cos(flip_angle)
    sina = torch.sin(flip_angle)
    cosa2 = (cosa + 1) / 2
    sina2 = 1 - cosa2
    ejp = torch.polar(torch.ones_like(phase), phase)
    inv_ejp = 1 / ejp
    # we need to stack the same dtype. +0j does not work in torchscript on cuda
    cosa2_complex = torch.complex(cosa2, torch.zeros_like(cosa2))
    cosa_complex = torch.complex(cosa, torch.zeros_like(cosa))
    new_shape = flip_angle.shape + (3, 3)  # noqa: RUF005 # not supported in torchscript

    return torch.stack(
        [
            cosa2_complex,
            ejp**2 * sina2,
            -1.0j * ejp * sina,
            inv_ejp**2 * sina2,
            cosa2_complex,
            1.0j * inv_ejp * sina,
            -0.5j * inv_ejp * sina,
            0.5j * ejp * sina,
            cosa_complex,
        ],
        -1,
    ).reshape(new_shape)


def rf(state: torch.Tensor, matrix: torch.Tensor) -> torch.Tensor:
    """Propagate EPG states through an RF rotation.

    Parameters
    ----------
    state
        EPG configuration states. Shape `(..., 3 (f_plus, f_minus, z), n)`
    matrix
        Rotation matrix describing the mixing of the EPG configuration states due to an RF pulse.
        Shape `(..., 3, 3)`

    Returns
    -------
        EPG configuration states after RF pulse. Shape `(..., 3 (f_plus, f_minus, z), n)`
    """
    return matrix.to(state) @ state


def gradient_dephasing(state: torch.Tensor) -> torch.Tensor:
    """Propagate EPG states through a "unit" gradient.

    Parameters
    ----------
    state
        EPG configuration states. Shape `(..., 3 (f_plus, f_minus, z), n)`
        with n being the number of configuration states > 1

    Returns
    -------
        EPG configuration states after gradient. Shape `(..., 3 (f_plus, f_minus, z), n)`
    """
    zero = state.new_zeros(state.shape[:-2] + (1,))  # noqa: RUF005 # not supported in torchscript
    f_plus = torch.cat((state[..., 1, 1:2].conj(), state[..., 0, :-1]), dim=-1)
    f_minus = torch.cat((state[..., 1, 1:], zero), -1)
    z = state[..., 2, :]
    return torch.stack((f_plus, f_minus, z), -2)


def relax_matrix(relaxation_time: torch.Tensor, t1: torch.Tensor, t2: torch.Tensor) -> torch.Tensor:
    """Calculate relaxation vector.

    Parameters
    ----------
    relaxation_time
        relaxation time
    t1
        longitudinal relaxation time
    t2
        transversal relaxation time

    Returns
    -------
        vector describing relaxation
    """
    e2 = torch.exp(-relaxation_time / t2)
    e1 = torch.exp(-relaxation_time / t1)
    e1, e2 = torch.broadcast_tensors(e1, e2)
    return torch.stack([e2, e2, e1], dim=-1)


def relax(states: torch.Tensor, relaxation_matrix: torch.Tensor, t1_recovery: bool = True) -> torch.Tensor:
    """Propagate EPG states through a period of relaxation and recovery.

    Parameters
    ----------
    states
        EPG configuration states. Shape `(..., 3 (f_plus, f_minus, z), n)`
    relaxation_matrix
        matrix describing EPG relaxation
    t1_recovery
        recovery of longitudinal EPG states

    Returns
    -------
        EPG configuration states after relaxation and recovery.
        Shape `(..., 3 (f_plus, f_minus, z), n)`
    """
    relaxation_matrix = relaxation_matrix.to(states)
    states = relaxation_matrix[..., None] * states

    if t1_recovery:
        states[..., 2, 0] = states[..., 2, 0] + (1 - relaxation_matrix[..., -1])
    return states


def acquisition(state: torch.Tensor, m0: torch.Tensor) -> torch.Tensor:
    """Calculate the acquired signal from the EPG state."""
    return m0 * state[..., 0, 0]


def initial_state(
    shape: Sequence[int], n_states: int = 20, device: torch.device | str = 'cpu', dtype: torch.dtype = torch.complex64
) -> torch.Tensor:
    """Generate initial EPG states.

    Parameters
    ----------
    shape
        Shape of the state tensor, excluding the last two dimensions. Should match the parameters shape.
    n_states
        Number of EPG configuration states.
    device
        Device to create the state tensor on.
    dtype
        Data type of the state tensor.

    Returns
    -------
        Initial EPG state tensor. Shape `(*shape, 3 (f_plus, f_minus, z), n)`
    """
    if n_states < 2:
        raise ValueError('Number of states should be at least 2.')
    states = torch.zeros(*shape, 3, n_states, device=device, dtype=dtype)
    # Start in equilibrium state
    states[..., :, 0] = torch.tensor((0.0, 0, 1.0))
    return states


class EPGBlock(TensorAttributeMixin, ABC):
    """Base class for EPG blocks."""

    def __call__(
        self, parameters: Parameters, states: torch.Tensor | int = 20
    ) -> tuple[torch.Tensor, tuple[torch.Tensor, ...]]:
        """Apply the block.

        Parameters
        ----------
        parameters
            Tissue parameters
        states
            EPG configuration states.
            If an integer value, the equilibrium state (0, 0, 1) will be initialized with
            the given number of EPG configuration states. The number should be large enough to capture the
            signal dynamics. More states increase the accuracy of the simulation but also the computational cost.
        """
        if isinstance(states, int):
            states = initial_state(
                parameters.shape, n_states=states, device=parameters.device, dtype=parameters.dtype.to_complex()
            )
        return super().__call__(parameters, states)

    @abstractmethod
    def forward(self, parameters: Parameters, states: torch.Tensor) -> tuple[torch.Tensor, tuple[torch.Tensor, ...]]:
        """Apply the block."""
        raise NotImplementedError

    @property
    def duration(self) -> torch.Tensor:
        """Duration of the block."""
        raise NotImplementedError


class RFBlock(EPGBlock):
    """RF pulse block."""

    def __init__(self, flip_angle: torch.Tensor | float, phase: torch.Tensor | float) -> None:
        """Initialize RF pulse block.

        Parameters
        ----------
        flip_angle
            flip angle in rad
        phase
            initial rf phase in rad
        """
        super().__init__()
        self.flip_angle = torch.as_tensor(flip_angle)
        self.phase = torch.as_tensor(phase)

    def forward(
        self,
        parameters: Parameters,
        states: torch.Tensor,
    ) -> tuple[torch.Tensor, tuple[torch.Tensor, ...]]:
        """Apply the RF pulse to the EPG state.

        Parameters
        ----------
        parameters
            Tissue parameters
        states
            EPG configuration states

        Returns
        -------
            EPG configuration states after the block and an empty list
        """
        matrix = rf_matrix(self.flip_angle, self.phase, parameters.relative_b1)
        return rf(states, matrix), ()

    @property
    def duration(self) -> torch.Tensor:
        """Duration of the block."""
        return torch.tensor(0.0)


class AcquisitionBlock(EPGBlock):
    """Acquisition block."""

    def forward(
        self,
        parameters: Parameters,
        states: torch.Tensor,
    ) -> tuple[torch.Tensor, tuple[torch.Tensor, ...]]:
        """Apply the acquisition block to the EPG state.

        Parameters
        ----------
        parameters
            Tissue parameters
        states
            EPG configuration states

        Returns
        -------
            EPG configuration states (unchanged) after the block and the acquired signal
        """
        return states, (acquisition(states, parameters.m0),)

    @property
    def duration(self) -> torch.Tensor:
        """Duration of the block."""
        return torch.tensor(0.0)


class GradientDephasingBlock(EPGBlock):
    """Gradient dephasing block."""

    def forward(
        self,
        parameters: Parameters,  # noqa: ARG002
        states: torch.Tensor,
    ) -> tuple[torch.Tensor, tuple[torch.Tensor, ...]]:
        """Apply the gradient dephasing block to the EPG state.

        Parameters
        ----------
        parameters
            Tissue parameters
        states
            EPG configuration states


        Returns
        -------
            EPG configuration states after the block and an empty list
        """
        return gradient_dephasing(states), ()

    @property
    def duration(self) -> torch.Tensor:
        """Duration of the block."""
        return torch.tensor(0.0)


class FispBlock(EPGBlock):
    """FISP data acquisition block.

    The block consists of a series of RF pulses with different flip angles and phases followed by an acquisition:
        - rf with flip_angle[i] and rf_phase[i]
        - relax for te[i]
        - acquisition
        - gradient_dephasing
        - relax for tr[i] - te[i]
    """

    flip_angles: torch.Tensor
    rf_phases: torch.Tensor
    te: torch.Tensor
    tr: torch.Tensor

    def __init__(
        self,
        flip_angles: torch.Tensor | float,
        rf_phases: torch.Tensor | float,
        te: torch.Tensor | float,
        tr: torch.Tensor | float,
    ) -> None:
        """Initialize the FISP block.

        Parameters
        ----------
        flip_angles
            Flip angles of the RF pulses in rad
        rf_phases
            Phase of the RF pulses in rad
        te
            Echo time
        tr
            Repetition time
        """
        super().__init__()

        flip_angles_, rf_phases_, te_, tr_ = map(torch.as_tensor, (flip_angles, rf_phases, te, tr))
        try:
            self.flip_angles, self.rf_phases, self.te, self.tr = torch.broadcast_tensors(
                flip_angles_, rf_phases_, te_, tr_
            )
        except RuntimeError:
            raise ValueError(
                f'Shapes of flip_angles ({flip_angles_.shape}), rf_phases ({rf_phases_.shape}), te ({te_.shape}) and '
                f'tr ({tr_.shape}) cannot be broadcasted.',
            ) from None

        self.flip_angles, self.rf_phases, self.te, self.tr = move_tensors_to_gpu_if_any_on_gpu(
            self.flip_angles, self.rf_phases, self.te, self.tr
        )

        if (self.te > self.tr).any():
            raise ValueError(f'echotime ({self.te}) should be smaller than repetition time ({self.tr}).')
        if (self.te < 0).any():
            raise ValueError(f'Negative echo time ({self.te.amin()}) not allowed.')

    @property
    def duration(self) -> torch.Tensor:
        """Duration of the block."""
        return self.tr.sum(0)

    def forward(
        self,
        parameters: Parameters,
        states: torch.Tensor,
    ) -> tuple[torch.Tensor, tuple[torch.Tensor, ...]]:
        """Apply the FISP block to the EPG state.

        Parameters
        ----------
        parameters
            Tissue parameters
        states
            EPG configuration states

        Returns
        -------
            EPG configuration states after the block and the acquired signals
        """
        signal = []
        # +1 for time dimension
        unsqueezed = unsqueeze_tensors_right(
            self.flip_angles, self.rf_phases, self.te, self.tr, ndim=parameters.ndim + 1
        )
        for flip_angle, rf_phase, te, tr in zip(*unsqueezed, strict=True):
            states = rf(states, rf_matrix(flip_angle, rf_phase, parameters.relative_b1))
            states = relax(states, relax_matrix(te, parameters.t1, parameters.t2))
            signal.append(acquisition(states, parameters.m0))
            states = gradient_dephasing(states)
            states = relax(states, relax_matrix((tr - te), parameters.t1, parameters.t2))
        return states, tuple(signal)


class TseBlock(EPGBlock):
    """TSE data acquisition block.

    The block consists of a 90° excitation pulse followed by a series of refocusing pulses with different flip angles
    and phases followed by an acquisition:
        - rf excitation with 90° and phase 90°
        - relax for te/2
        - gradient_dephasing
        - rf with flip_angle[i] and rf_phase[i]
        - gradient_dephasing
        - relax for te/2
        - acquisition
    """

    refocusing_flip_angles: torch.Tensor
    refocusing_rf_phases: torch.Tensor
    te: torch.Tensor

    def __init__(
        self,
        refocusing_flip_angles: torch.Tensor | float,
        refocusing_rf_phases: torch.Tensor | float,
        te: float,
    ) -> None:
        """Initialize the TSE block.

        Parameters
        ----------
        refocusing_flip_angles
            Flip angles of the refocusing RF pulses in rad
        refocusing_rf_phases
            Phase of the refocusing RF pulses in rad
        te
            Echo time
        """
        super().__init__()

        refocusing_flip_angles_, refocusing_rf_phases_ = map(
            torch.as_tensor, (refocusing_flip_angles, refocusing_rf_phases)
        )
        try:
            self.refocusing_flip_angles, self.refocusing_rf_phases = torch.broadcast_tensors(
                refocusing_flip_angles_, refocusing_rf_phases_
            )
        except RuntimeError:
            raise ValueError(
                f'Shapes of flip_angles ({refocusing_flip_angles_.shape}) and rf_phases ({refocusing_rf_phases_.shape})'
                f' cannot be broadcasted.',
            ) from None

        self.te = torch.as_tensor(te)
        if (self.te < 0).any():
            raise ValueError(f'Negative echo time ({self.te.amin()}) not allowed.')

    @property
    def duration(self) -> torch.Tensor:
        """Duration of the block."""
        return self.te * self.refocusing_flip_angles.shape[0]

    def forward(
        self,
        parameters: Parameters,
        states: torch.Tensor,
    ) -> tuple[torch.Tensor, tuple[torch.Tensor, ...]]:
        """Apply the TSE block to the EPG state.

        Parameters
        ----------
        parameters
            Tissue parameters
        states
            EPG configuration states

        Returns
        -------
            EPG configuration states after the block and the acquired signals
        """
        signal = []
        # +1 for time dimension
        unsqueezed = unsqueeze_tensors_right(
            self.refocusing_flip_angles, self.refocusing_rf_phases, ndim=parameters.ndim + 1
        )
        states = rf(states, rf_matrix(torch.tensor(torch.pi / 2), torch.tensor(torch.pi / 2), parameters.relative_b1))
        for refocusing_flip_angle, refocusing_rf_phase in zip(*unsqueezed, strict=True):
            states = relax(states, relax_matrix(self.te / 2, parameters.t1, parameters.t2))
            states = gradient_dephasing(states)
            states = rf(states, rf_matrix(refocusing_flip_angle, refocusing_rf_phase, parameters.relative_b1))
            states = gradient_dephasing(states)
            states = relax(states, relax_matrix(self.te / 2, parameters.t1, parameters.t2))
            signal.append(acquisition(states, parameters.m0))
        return states, tuple(signal)


class InversionBlock(EPGBlock):
    """T1 Inversion Preparation Block.

    The inversion pulse is assumed to be B1 insensitive.
    """

    def __init__(self, inversion_time: torch.Tensor | float) -> None:
        """Initialize the inversion block.

        Parameters
        ----------
        inversion_time
            Inversion time
        """
        super().__init__()

        self.inversion_time = torch.as_tensor(inversion_time)
        if (self.inversion_time < 0).any():
            raise ValueError(f'Negative inversion time ({self.inversion_time.amin()}) not allowed.')

    def forward(
        self,
        parameters: Parameters,
        states: torch.Tensor,
    ) -> tuple[torch.Tensor, tuple[torch.Tensor, ...]]:
        """Apply the inversion block to the EPG state.

        Parameters
        ----------
        parameters
            Tissue parameters
        states
            EPG configuration states

        Returns
        -------
            EPG configuration states after the block and an empty list
        """
        states = rf(states, rf_matrix(torch.tensor(torch.pi), torch.tensor(0.0)))
        states = relax(states, relax_matrix(self.inversion_time, parameters.t1, parameters.t2))
        return states, ()

    @property
    def duration(self) -> torch.Tensor:
        """Duration of the block."""
        return self.inversion_time


class T2PrepBlock(EPGBlock):
    """T2 Preparation Block.

    Consists of a 90° pulse, relaxation during TE/2, 180° pulse, relaxation during TE/2 and a -90° pulse.

    The pulses are assumed to be B1 insensitive.
    """

    def __init__(self, te: torch.Tensor | float) -> None:
        """Initialize the T2 preparation block.

        Parameters
        ----------
        te
            Echo time of the T2 preparation block
        """
        super().__init__()

        self.te = torch.as_tensor(te)
        if (self.te < 0).any():
            raise ValueError(f'Negative echo time ({self.te.amin()}) not allowed.')

    def forward(
        self,
        parameters: Parameters,
        states: torch.Tensor,
    ) -> tuple[torch.Tensor, tuple[torch.Tensor, ...]]:
        """Apply the T2 preparation block to the EPG state.

        Parameters
        ----------
        parameters
            Tissue parameters
        states
            EPG configuration states

        Returns
        -------
            EPG configuration states after the block and an empty list
        """
        # 90° pulse -> relaxation during TE/2 -> 180° pulse -> relaxation during TE/2 -> -90° pulse
        te2_relax = relax_matrix(self.te / 2, parameters.t1, parameters.t2)
        states = rf(states, rf_matrix(torch.tensor(torch.pi / 2), torch.tensor(0.0)))
        states = relax(states, te2_relax)
        states = rf(states, rf_matrix(torch.tensor(torch.pi), torch.tensor(torch.pi / 2)))
        states = relax(states, te2_relax)
        states = rf(states, rf_matrix(torch.tensor(torch.pi / 2), -torch.tensor(torch.pi)))
        # Spoiler
        states = gradient_dephasing(states)
        return states, ()

    @property
    def duration(self) -> torch.Tensor:
        """Duration of the block."""
        return self.te


class T1RhoPrepBlock(EPGBlock):
    """T1 Rho Preparation Block.

    Consists of a 90° pulse, then T1rho relaxation during a spin-lock pulse followed by a -90° pulse.

    The pulses are assumed to be B1 insensitive.
    """

    def __init__(self, spin_lock_duration: torch.Tensor | float) -> None:
        """Initialize the T1 Rho preparation block.

        Parameters
        ----------
        spin_lock_duration
            Duration of the spin-lock pulse
        """
        super().__init__()

        self.spin_lock_duration = torch.as_tensor(spin_lock_duration)

    def forward(
        self,
        parameters: Parameters,
        states: torch.Tensor,
    ) -> tuple[torch.Tensor, tuple[torch.Tensor, ...]]:
        """Apply the T1 Rho preparation block to the EPG state.

        Parameters
        ----------
        parameters
            Tissue parameters
        states
            EPG configuration states

        Returns
        -------
            EPG configuration states after the block and an empty list
        """
        if parameters.t1_rho is not None:
            # 90° pulse -> T1 rho relaxation during spin lock pulse -> -90° pulse
            states = rf(states, rf_matrix(torch.tensor(torch.pi / 2), torch.tensor(0.0)))
            e_t1_rho = torch.exp(-self.spin_lock_duration / parameters.t1_rho)
            t1_rho_relax_matrix = torch.stack([e_t1_rho, e_t1_rho, torch.ones_like(e_t1_rho)], dim=-1)
            states = relax(states, t1_rho_relax_matrix, t1_recovery=False)
            states = rf(states, rf_matrix(torch.tensor(torch.pi / 2), -torch.tensor(torch.pi)))
            # Spoiler
            states = gradient_dephasing(states)
        return states, ()

    @property
    def duration(self) -> torch.Tensor:
        """Duration of the block."""
        return self.spin_lock_duration


class DelayBlock(EPGBlock):
    """Delay Block."""

    def __init__(self, delay_time: torch.Tensor | float) -> None:
        """Initialize the delay block.

        Parameters
        ----------
        delay_time
            Delay time
        """
        super().__init__()
        self.delay_time = torch.as_tensor(delay_time)
        if (self.delay_time < 0).any():
            raise ValueError(f'Negative delay time ({self.delay_time.amin()})')

    def forward(
        self,
        parameters: Parameters,
        states: torch.Tensor,
    ) -> tuple[torch.Tensor, tuple[torch.Tensor, ...]]:
        """Apply the delay block to the EPG state.

        Parameters
        ----------
        parameters
            Tissue parameters
        states
            EPG configuration states

        Returns
        -------
            EPG configuration states after the block and an empty list
        """
        (delay_time,) = unsqueeze_tensors_right(self.delay_time, ndim=parameters.ndim)
        states = relax(states, relax_matrix(delay_time, parameters.t1, parameters.t2))
        return states, ()

    @property
    def duration(self) -> torch.Tensor:
        """Duration of the block."""
        return self.delay_time


class EPGSequence(torch.nn.ModuleList, EPGBlock):
    """Sequence of EPG blocks.

    A sequence as multiple blocks, such as preparation pulses, acquisition blocks and delays.

    Classic MRF with a single inversion-pulse at the beginning followed by a train of RF pulses with different flip
    angles as described in [MA2013]_ can be simulated by the following sequence:
     - InversionBlock
     - FispBlock with varying flip_angles

        You can also nest `EPGSequence`, i.e., use it as a block in another `EPGSequence`. This allows to describe
        complex MR acquisitions with multiple repetitions of the same blocks.


    References
    ----------
    .. [MA2013] Ma, D et al.(2013) Magnetic resonance fingerprinting. Nature 495 http://dx.doi.org/10.1038/nature11971


    """

    def __init__(self, blocks: Sequence[EPGBlock] = ()) -> None:
        """Initialize an EPG Sequence.

        Parameters
        ----------
        blocks
            blocks such as RF, delays, acquisitions etc.
        """
        torch.nn.ModuleList.__init__(self, blocks)

    def forward(
        self,
        parameters: Parameters,
        states: torch.Tensor,
    ) -> tuple[torch.Tensor, tuple[torch.Tensor, ...]]:
        """Apply the sequence of blocks to the EPG states.

        Parameters
        ----------
        parameters
            Tissue parameters
        states
            EPG configuration states.

        Returns
        -------
            EPG configuration states after the sequence of blocks and the acquired signals
        """
        signals: list[torch.Tensor] = []
        for block in self:
            assert isinstance(block, EPGBlock)  # mypy # noqa: S101
            states, signal = block(parameters, states)
            signals.extend(signal)
        return states, tuple(signals)

    @property
    def duration(self) -> torch.Tensor:
        """Duration of the block."""
        return sum((block.duration for block in self if isinstance(block, EPGBlock)), start=torch.tensor(0.0))


__all__ = [
    'AcquisitionBlock',
    'DelayBlock',
    'EPGSequence',
    'FispBlock',
    'GradientDephasingBlock',
    'InversionBlock',
    'Parameters',
    'RFBlock',
    'T1RhoPrepBlock',
    'T2PrepBlock',
    'TseBlock',
    'initial_state',
]



"""Functions for tensor shaping, unit conversion, typing, etc."""

from mrpro.utils import slice_profiles
from mrpro.utils import typing
from mrpro.utils import unit_conversion
from mrpro.utils.fill_range import fill_range_
from mrpro.utils.smap import smap
from mrpro.utils.reduce_repeat import reduce_repeat
from mrpro.utils.indexing import Indexer, normalize_index
from mrpro.utils.pad_or_crop import pad_or_crop
from mrpro.utils.split_idx import split_idx
from mrpro.utils.sliding_window import sliding_window
from mrpro.utils.summarize import summarize_object, summarize_values
from mrpro.utils.reshape import broadcast_right, broadcasted_rearrange, unsqueeze_left, unsqueeze_right, reduce_view, reshape_broadcasted, ravel_multi_index, unsqueeze_tensors_left, unsqueeze_tensors_right, unsqueeze_at, unsqueeze_tensors_at, broadcasted_concatenate
from mrpro.utils.TensorAttributeMixin import TensorAttributeMixin
from mrpro.utils.interpolate import interpolate, apply_lowres
from mrpro.utils.RandomGenerator import RandomGenerator

__all__ = [
    "Indexer",
    "RandomGenerator",
    "TensorAttributeMixin",
    "apply_lowres",
    "broadcast_right",
    "broadcasted_concatenate",
    "broadcasted_rearrange",
    "fill_range_",
    "interpolate",
    "normalize_index",
    "pad_or_crop",
    "ravel_multi_index",
    "reduce_repeat",
    "reduce_view",
    "reshape_broadcasted",
    "slice_profiles",
    "sliding_window",
    "smap",
    "split_idx",
    "summarize_object",
    "summarize_values",
    "typing",
    "unit_conversion",
    "unsqueeze_at",
    "unsqueeze_left",
    "unsqueeze_right",
    "unsqueeze_tensors_at",
    "unsqueeze_tensors_left",
    "unsqueeze_tensors_right"
]



"""Averaging operator."""

from collections.abc import Sequence
from warnings import warn

import torch

from mrpro.operators.LinearOperator import LinearOperator


class AveragingOp(LinearOperator):
    """Averaging operator.

    This operator averages the input tensor along a specified dimension.
    The averaging is performed over groups of elements defined by the `idx` parameter.
    The output tensor will have the same shape as the input tensor, except for the `dim` dimension,
    which will have a size equal to the number of groups specified in `idx`. For each group,
    the average of the elements in that group is computed.

    For example, this operator can be used to simulate the effect of a sliding window average
    on a signal model.
    """

    def __init__(
        self,
        dim: int,
        idx: Sequence[Sequence[int] | torch.Tensor | slice] | torch.Tensor = (slice(None),),
        domain_size: int | None = None,
    ) -> None:
        """Initialize the averaging operator.

        Parameters
        ----------
        dim
            The dimension along which to average.
        idx
            The indices of the input tensor to average over. Each element of the sequence will result in a
            separate entry in the `dim` dimension of the output tensor.
            The entries can be either a sequence of integers or an integer tensor, a slice object, or a boolean tensor.
        domain_size
            The size of the input along `dim`. It is only used in the `adjoint` method.
            If not set, the size will be guessed from the input tensor during the forward pass.
        """
        super().__init__()
        self.domain_size = domain_size
        self._last_domain_size = domain_size
        self.idx = idx
        self.dim = dim

    def __call__(self, x: torch.Tensor) -> tuple[torch.Tensor]:
        """Apply the averaging operator to the input tensor.

        Parameters
        ----------
        x
            Input tensor to be averaged.

        Returns
        -------
            Averaged tensor.
        """
        return super().__call__(x)

    def forward(self, x: torch.Tensor) -> tuple[torch.Tensor]:
        """Apply forward of AveragingOp.

        .. note::
            Prefer calling the instance of the AveragingOp operator as ``operator(x)`` over
            directly calling this method. See this PyTorch `discussion <https://discuss.pytorch.org/t/is-model-forward-x-the-same-as-model-call-x/33460/3>`_.
        """
        if self.domain_size and self.domain_size != x.shape[self.dim]:
            raise ValueError(f'Expected domain size {self.domain_size}, got {x.shape[self.dim]}')
        self._last_domain_size = x.shape[self.dim]

        placeholder = (slice(None),) * (self.dim % x.ndim)
        averaged = torch.stack([x[(*placeholder, i)].mean(self.dim) for i in self.idx], self.dim)
        return (averaged,)

    def adjoint(self, x: torch.Tensor) -> tuple[torch.Tensor]:
        """Apply the adjoint of the averaging operator.

        The adjoint operation distributes the values from the input tensor `x`
        (which corresponds to the output of the forward pass) back to a
        tensor matching the original domain size. Each element in the output
        of the adjoint receives a contribution from `x` divided by the
        number of elements in its original averaging group.

        Parameters
        ----------
        x
            Input tensor, corresponding to the output of the forward operation.

        Returns
        -------
            Tensor with values distributed back to the original domain.
        """
        if self.domain_size is None:
            if self._last_domain_size is None:
                raise ValueError('Domain size is not set. Please set it explicitly or run forward first.')
            warn(
                'Domain size is not set. Guessing the last used input size of the forward pass. '
                'Consider setting the domain size explicitly.',
                stacklevel=2,
            )
            self.domain_size = self._last_domain_size

        adjoint = x.new_zeros(*x.shape[: self.dim], self.domain_size, *x.shape[self.dim + 1 :])
        placeholder = (slice(None),) * (self.dim % x.ndim)
        for i, group in enumerate(self.idx):
            if isinstance(group, slice):
                n = len(range(*group.indices(self.domain_size)))
            elif isinstance(group, torch.Tensor) and group.dtype == torch.bool:
                n = int(group.sum())
            else:
                n = len(group)

            adjoint[(*placeholder, group)] += (
                x[(*placeholder, i, None)].expand(*x.shape[: self.dim], n, *x.shape[self.dim + 1 :]) / n
            )

        return (adjoint,)



"""Inati method for coil sensitivity map calculation."""

import torch
from einops import einsum

from mrpro.data.SpatialDimension import SpatialDimension
from mrpro.utils.sliding_window import sliding_window


def inati(
    coil_img: torch.Tensor,
    smoothing_width: SpatialDimension[int] | int,
) -> torch.Tensor:
    """Calculate a coil sensitivity map (csm) using the Inati method [INA2013]_ [INA2014]_.

    This is for a single set of coil images. The input should be a tensor with dimensions `(coils, z, y, x)`. The output
    will have the same dimensions. Either apply this function individually to each set of coil images, or see
    `~mrpro.data.CsmData.from_idata_inati` which performs this operation on a whole dataset.

    .. [INA2013] Inati S, Hansen M, Kellman P (2013) A solution to the phase problem in adaptvie coil combination.
       in Proceedings of the 21st Annual Meeting of ISMRM, Salt Lake City, USA, 2672.

    .. [INA2014] Inati S, Hansen M (2014) A Fast Optimal Method for Coil Sensitivity Estimation and Adaptive Coil
       Combination for Complex images. in Proceedings of Joint Annual Meeting ISMRM-ESMRMB, Milan, Italy, 7115.

    Parameters
    ----------
    coil_img
        images for each coil element
    smoothing_width
        Size of the smoothing kernel
    """
    # After 10 power iterations we will have a very good estimate of the singular vector
    n_power_iterations = 10
    eps = 1e-8  # for numerical stability

    if isinstance(smoothing_width, int):
        smoothing_width = SpatialDimension(
            z=smoothing_width if coil_img.shape[-3] > 1 else 1, y=smoothing_width, x=smoothing_width
        )

    if any(ks % 2 != 1 for ks in [smoothing_width.z, smoothing_width.y, smoothing_width.x]):
        raise ValueError('kernel_size must be odd')

    ks_halved = [ks // 2 for ks in smoothing_width.zyx]
    padded_coil_img = torch.nn.functional.pad(
        coil_img,
        (ks_halved[-1], ks_halved[-1], ks_halved[-2], ks_halved[-2], ks_halved[-3], ks_halved[-3]),
        mode='replicate',
    )
    # Get the voxels in an ROI defined by the smoothing_width around each voxel leading to shape
    # (z y x coils window=prod(smoothing_width))
    coil_img_roi = sliding_window(padded_coil_img, smoothing_width.zyx, dim=(-3, -2, -1)).flatten(-3)
    coil_img_cov = einsum(
        coil_img_roi.conj(),
        coil_img_roi,
        '... coils1 window,... coils2 window->... coils1 coils2',
    )

    singular_vector = torch.sum(coil_img_roi, dim=-1)  # z y x coils
    singular_vector /= singular_vector.norm(dim=-1, keepdim=True) + eps
    for _ in range(n_power_iterations):
        singular_vector = einsum(coil_img_cov, singular_vector, '... coils1 coils2,... coils2->... coils1')
        singular_vector /= singular_vector.norm(dim=-1, keepdim=True) + eps

    singular_value = einsum(coil_img_roi, singular_vector, '... coils window,... coils->... window')
    phase = singular_value.sum(-1)
    phase /= phase.abs() + eps
    csm = einsum(singular_vector.conj(), phase, '... coils,...->coils ...')  # coils z y x
    return csm



"""Operator returning the phase of the input."""

import torch

from mrpro.operators.EndomorphOperator import EndomorphOperator, endomorph


class PhaseOp(EndomorphOperator):
    """Phase of input tensors."""

    @endomorph
    def __call__(self, *x: torch.Tensor) -> tuple[torch.Tensor, ...]:
        """Calculate the element-wise phase (angle) of input tensors.

        Parameters
        ----------
        *x
            One or more input tensors. Can be complex or real.

        Returns
        -------
            A tuple of tensors, where each tensor contains the element-wise
            phase of the corresponding input tensor. The phase is in radians.
        """
        return super().__call__(*x)

    @endomorph
    def forward(self, *x: torch.Tensor) -> tuple[torch.Tensor, ...]:
        """Apply forward of PhaseOp.

        .. note::
            Prefer calling the instance of the PhaseOp operator as ``operator(x)`` over
            directly calling this method. See this PyTorch `discussion <https://discuss.pytorch.org/t/is-model-forward-x-the-same-as-model-call-x/33460/3>`_.
        """
        return tuple([torch.angle(xi) for xi in x])



"""Spatial and temporal filters."""

import warnings
from collections.abc import Sequence
from functools import reduce
from math import ceil
from typing import Literal

import numpy as np
import torch
from einops import repeat


def filter_separable(
    x: torch.Tensor,
    kernels: Sequence[torch.Tensor],
    dim: Sequence[int],
    pad_mode: Literal['constant', 'reflect', 'replicate', 'circular', 'none'] = 'constant',
    pad_value: float = 0.0,
) -> torch.Tensor:
    """Apply the separable filter kernels to the tensor x along the axes dim.

    Does padding to keep the output the same size as the input.

    Parameters
    ----------
    x
        Tensor to filter
    kernels
        List of 1D kernels to apply to the tensor x
    dim
        Axes to filter over. Must have the same length as kernels.
    pad_mode
        Padding mode
    pad_value
        Padding value for pad_mode = constant

    Returns
    -------
    The filtered tensor, with the same shape as the input unless pad_mode is 'none' and
    and promoted dtype of the input and the kernels.
    """
    if len(dim) != len(kernels):
        raise ValueError('Must provide matching length kernels and dim arguments.')

    # normalize dim to allow negative indexing in input
    dim = tuple([a % x.ndim for a in dim])
    if len(dim) != len(set(dim)):
        raise ValueError(f'Dim must be unique. Normalized dims are {dim}')

    if pad_mode == 'constant' and pad_value == 0:
        # padding is done inside the convolution
        padding_conv = 'same'
    else:
        # padding is done with pad() before the convolution
        padding_conv = 'valid'

    # output will be of the promoted type of the input and the kernels
    target_dtype = reduce(torch.promote_types, [k.dtype for k in kernels], x.dtype)
    x = x.to(target_dtype)

    for kernel, d in zip(kernels, dim, strict=False):
        kernel = kernel.to(device=x.device, dtype=target_dtype)
        # moveaxis is not implemented for batched tensors, so vmap would fail.
        # thus we use permute.
        idx = list(range(x.ndim))
        # swapping the last axis and the axis to filter over
        idx[d], idx[-1] = idx[-1], idx[d]
        x = x.permute(idx)
        # flatten first to allow for circular, replicate and reflection padding for arbitrary tensor size
        x_flat = x.flatten(end_dim=-2)
        if padding_conv == 'valid' and pad_mode != 'none':
            left_pad = (len(kernel) - 1) // 2
            right_pad = (len(kernel) - 1) - left_pad
            x_flat = torch.nn.functional.pad(x_flat, pad=(left_pad, right_pad), mode=pad_mode, value=pad_value)
        x = torch.nn.functional.conv1d(
            repeat(x_flat, 'batch x -> batch channels x', channels=1),
            repeat(kernel, 'x -> batch channels x', batch=1, channels=1),
            padding=padding_conv,
        ).reshape(*x.shape[:-1], -1)
        # for a single permutation, this undoes the permutation
        x = x.permute(idx)
    return x


def gaussian_filter(
    x: torch.Tensor,
    sigmas: float | Sequence[float] | torch.Tensor,
    dim: int | Sequence[int] | None = None,
    truncate: int = 3,
    pad_mode: Literal['constant', 'reflect', 'replicate', 'circular'] = 'constant',
    pad_value: float = 0.0,
) -> torch.Tensor:
    """Apply a and-Gaussian filter.

    Parameters
    ----------
    x
        Tensor to filter
    sigmas
        Standard deviation for Gaussian kernel. If iterable, must have length equal to the number of axes.
    dim
        Axis or axes to filter over. If None, filters over all axes.
    truncate
        Truncate the filter at this many standard deviations.
    pad_mode
        Padding mode
    pad_value
        Padding value for pad_mode = constant
    """
    if dim is None:
        dim = tuple(range(x.ndim))
    elif isinstance(dim, int):
        dim = (dim,)
    sigmas = torch.as_tensor(sigmas) if np.iterable(sigmas) else torch.tensor([sigmas] * len(dim))
    if not torch.all(sigmas > 0):
        raise ValueError('`sigmas` must be positive')

    if len(sigmas) != len(dim):
        raise ValueError('Must provide matching length sigmas and dim arguments. ')

    kernels = tuple(
        [
            torch.exp(-0.5 * (torch.arange(-ceil(truncate * sigma), ceil(truncate * sigma) + 1) / sigma) ** 2)
            for sigma in sigmas
        ]
    )
    kernels = tuple([(k / k.sum()).to(device=x.device) for k in kernels])
    x_filtered = filter_separable(x, kernels, dim, pad_mode, pad_value)
    return x_filtered


def uniform_filter(
    x: torch.Tensor,
    width: int | Sequence[int] | torch.Tensor,
    dim: int | Sequence[int] | None = None,
    pad_mode: Literal['constant', 'reflect', 'replicate', 'circular'] = 'constant',
    pad_value: float = 0.0,
) -> torch.Tensor:
    """Apply a and-uniform filter.

    Parameters
    ----------
    x
        Tensor to filter
    width
        Width of uniform kernel. If iterable, must have length equal to the number of axes.
    dim
        Axis or axes to filter over. If None, filters over all axes.
    pad_mode
        Padding mode
    pad_value
        Padding value for pad_mode = constant
    """
    if dim is None:
        dim = tuple(range(x.ndim))
    elif isinstance(dim, int):
        dim = (dim,)
    width = torch.as_tensor(width) if np.iterable(width) else torch.tensor([width] * len(dim))
    if not torch.all(width > 0):
        raise ValueError('width must be positive.')
    if torch.any(width % 2 != 1):
        warnings.warn('width should be odd.', stacklevel=2)
    if len(width) != len(dim):
        raise ValueError('Must provide matching length width and dim arguments. ')
    width = torch.minimum(width, torch.tensor(x.shape)[(dim), ...])

    kernels = tuple([torch.ones(width, device=x.device) / width for width in width])
    x_filtered = filter_separable(x, kernels, dim, pad_mode, pad_value)
    return x_filtered



"""Index for splitting data."""

import torch


def split_idx(idx: torch.Tensor, blocksize: int, overlap: int = 0, cyclic: bool = False) -> torch.Tensor:
    """Split a tensor of indices into different blocks.

    Parameters
    ----------
    idx
        1D indices to be split into different blocks.
    blocksize
        Number of points per block.
    overlap
        Number of points overlapping between blocks, default of ``0`` means no overlap between blocks.
    cyclic
        Last block is filled up with points from the first block, e.g. due to cyclic cardiac motion.


    Example:
    # idx = [1,2,3,4,5,6,7,8,9], np_per_block = 5, np_overlap = 3, cycle = True
    #
    # idx:     1 2 3 4 5 6 7 8 9
    # block 0: 0 0 0 0 0
    # block 1:     1 1 1 1 1
    # block 2:         2 2 2 2 2
    # block 3: 3 3         3 3 3

    Returns
    -------
        2D indices to split data into different blocks in the shape `[block, index]`.

    Raises
    ------
    `ValueError`
        If the provided idx is not 1D.
    `ValueError`
        If the overlap is smaller than the number of points per block.
    """
    # Make sure idx is 1D
    if idx.ndim != 1:
        raise ValueError('idx should be a 1D vector.')

    # Make sure overlap is not larger than the number of points in a block
    if overlap >= blocksize:
        raise ValueError('Overlap has to be smaller than the number of points in a block.')

    # Calculate number of blocks
    # 1 2 3 4 5 6 7 8 9
    # x x                       step
    #     x x x                 np_overlap
    # x x x x x                 np_per_block
    step = blocksize - overlap

    # For cyclic splitting utilize beginning of index to maximize number of blocks
    if cyclic:
        idx = torch.concat((idx, idx[:step]))

    return idx.unfold(dimension=0, size=blocksize, step=step)



"""Mixin for smarter tensor attributes."""

from typing import Any

import torch


class TensorAttributeMixin(torch.nn.Module):
    """Create tensor attributes as buffer."""

    def __setattr__(self, name: str, value: Any) -> None:  # noqa: ANN401
        """Set attribute.

        Set tensors not requiring gradients as buffer.

        Parameters
        ----------
        name
            name of the attribute.
        value
            attribute to set.
        """
        if isinstance(value, torch.Tensor) and not isinstance(value, torch.nn.Parameter) and not value.requires_grad:
            self.register_buffer(name, value)
        else:
            super().__setattr__(name, value)



"""KTrajectory dataclass."""

from collections.abc import Callable
from typing import Literal

import ismrmrd
import numpy as np
import torch
from typing_extensions import Self

from mrpro.data.Dataclass import Dataclass
from mrpro.data.enums import TrajType
from mrpro.data.SpatialDimension import SpatialDimension
from mrpro.utils.reduce_repeat import reduce_repeat
from mrpro.utils.reshape import unsqueeze_at
from mrpro.utils.typing import FileOrPath


class KTrajectory(Dataclass):
    """K-space trajectory.

    Contains the trajectory in k-space along the three dimensions `kz`, `ky`, `kx`,
    i.e. describes where in k-space each data point was acquired.

    The shape of each of `kx`, `ky`, `kz` is `(*other, coils=1, k2, k1, k0)`,
    where `other` can span multiple dimensions.

    Example for 2D-Cartesian trajectories:

        - `kx` changes along `k0` and is frequency encoding,
        - `ky` changes along `k1` and is phase encoding
        - `kz` is zero with shape `(1, 1, 1, 1, 1)`
    """

    kz: torch.Tensor
    """Trajectory in z direction / phase encoding direction k2 if Cartesian. Shape `(*other, coils=1, k2, k1, k0)`"""

    ky: torch.Tensor
    """Trajectory in y direction / phase encoding direction k1 if Cartesian. Shape `(*other, coils=1, k2, k1, k0)`"""

    kx: torch.Tensor
    """Trajectory in x direction / phase encoding direction k0 if Cartesian. Shape `(*other, coils=1, k2, k1, k0)`"""

    grid_detection_tolerance: float = 1e-3
    """tolerance of how close trajectory positions have to be to integer grid points."""

    repeat_detection_tolerance: float | None = 1e-3
    """tolerance for repeat detection. Set to `None` to disable."""

    def __post_init__(self) -> None:
        """Reduce repeated dimensions to singletons."""

        def as_any_float(tensor: torch.Tensor) -> torch.Tensor:
            return tensor.float() if not tensor.is_floating_point() else tensor

        if self.repeat_detection_tolerance is not None:
            kz, ky, kx = (
                as_any_float(reduce_repeat(tensor, self.repeat_detection_tolerance))
                for tensor in (self.kz, self.ky, self.kx)
            )
            # use of setattr due to frozen dataclass
            object.__setattr__(self, 'kz', kz)
            object.__setattr__(self, 'ky', ky)
            object.__setattr__(self, 'kx', kx)

        try:
            shape = self.shape
        except RuntimeError:
            raise ValueError('The k-space trajectory dimensions must be broadcastable.') from None

        if len(shape) < 5:
            raise ValueError('The k-space trajectory tensors should each have at least 5 dimensions.')

    @classmethod
    def from_tensor(
        cls,
        tensor: torch.Tensor,
        stack_dim: int = 0,
        axes_order: Literal['zxy', 'zyx', 'yxz', 'yzx', 'xyz', 'xzy'] = 'zyx',
        repeat_detection_tolerance: float | None = 1e-6,
        grid_detection_tolerance: float = 1e-3,
        scaling_matrix: SpatialDimension | None = None,
    ) -> Self:
        """Create a KTrajectory from a tensor representation of the trajectory.

        Reduces repeated dimensions to singletons if repeat_detection_tolerance is not set to `None`.

        Parameters
        ----------
        tensor
            The tensor representation of the trajectory.
            This should be a 5-dim tensor, with (`kz`, `ky`, `kx`) stacked in this order along `stack_dim`.
        stack_dim
            The dimension in the tensor along which the directions are stacked.
        axes_order
            The order of the axes in the tensor. The MRpro convention is 'zyx'.
        repeat_detection_tolerance
            Tolerance for detecting repeated dimensions (broadcasting).
            If trajectory points differ by less than this value, they are considered identical.
            Set to None to disable this feature.
        grid_detection_tolerance
            Tolerance for detecting whether trajectory points align with integer grid positions.
            This tolerance is applied after rescaling if `scaling_matrix` is provided.
        scaling_matrix
            If a scaling matrix is provided, the trajectory is rescaled to fit within
            the dimensions of the matrix. If not provided, the trajectory remains unchanged.

        """
        ks = tensor.unbind(dim=stack_dim)
        kz, ky, kx = (ks[axes_order.index(axis)] for axis in 'zyx')

        def rescale(k: torch.Tensor, size: float) -> torch.Tensor:
            max_abs_range = 2 * k.abs().max()
            if size < 2 or max_abs_range < 1e-6:
                # a single encoding point should be at zero
                # avoid division by zero
                return torch.zeros_like(k)
            return k * (size / max_abs_range)

        if scaling_matrix is not None:
            kz = rescale(kz, scaling_matrix.z)
            ky = rescale(ky, scaling_matrix.y)
            kx = rescale(kx, scaling_matrix.x)

        return cls(
            kz,
            ky,
            kx,
            repeat_detection_tolerance=repeat_detection_tolerance,
            grid_detection_tolerance=grid_detection_tolerance,
        )

    @classmethod
    def from_ismrmrd(
        cls,
        filename: FileOrPath,
        dataset_idx: int = -1,
        acquisition_filter_criterion: Callable[[ismrmrd.Acquisition], bool] | None = None,
        normalize: bool = False,
    ) -> Self:
        """Read a k-space trajectory from an ISMRMRD file.

        The trajectory is extracted from an ISMRMRD file. If the encoding matrix is set in the header,
        the trajectory is rescaled to fit within the dimensions of the matrix.

        Parameters
        ----------
        filename
            The path to the ISMRMRD file.
        dataset_idx
            The index of the dataset in the file.
        acquisition_filter_criterion
            A function that takes an ISMRMRD acquisition and returns a boolean.
            If provided, only acquisitions for which the function returns `True` are included in the trajectory.
            `None` includes all acquisitions.
        normalize
            Normalize the trajectory to the encoding matrix. Consider enabling for non-cartesian trajectories or
            non-normalized ISMRMRD trajectories.
        """
        with ismrmrd.File(filename, 'r') as file:
            datasets = list(file.keys())
            if not datasets:
                raise ValueError('No datasets found in the ISMRMRD file.')
            if not -len(datasets) <= dataset_idx < len(datasets):
                raise ValueError(f'Dataset index {dataset_idx} out of range, available datasets: {datasets}')
            dataset = file[datasets[dataset_idx]]

            acquisitions = [
                acq
                for acq in dataset.acquisitions
                if acquisition_filter_criterion is None or acquisition_filter_criterion(acq)
            ]
            if not acquisitions:
                raise ValueError('No matching acquisitions found in the ISMRMRD file.')
            traj = torch.stack([torch.as_tensor(acq.traj, dtype=torch.float32) for acq in acquisitions], dim=0)
            if not normalize:
                scaling_matrix = None
            elif (
                dataset.header.encoding
                and dataset.header.encoding[0].encodedSpace
                and dataset.header.encoding[0].encodedSpace.matrixSize
            ):
                scaling_matrix = SpatialDimension.from_xyz(dataset.header.encoding[0].encodedSpace.matrixSize)
            else:
                raise ValueError(
                    'Requested normalization, but the ISMRMRD file does not contain an encoding matrix size. '
                    'Consider adding it to the header.'
                )

        if traj.shape[-1] != 3:  # enforce 3D trajectory
            zero = torch.zeros_like(traj[..., :1])
            traj = torch.cat([traj, *([zero] * (3 - traj.shape[-1]))], dim=-1)
        traj = unsqueeze_at(traj, dim=-3, n=5 - traj.ndim + 1)  # +1 due to stack_dim

        return cls.from_tensor(traj, stack_dim=-1, axes_order='xyz', scaling_matrix=scaling_matrix)

    @property
    def type_along_kzyx(self) -> tuple[TrajType, TrajType, TrajType]:
        """Type of trajectory along kz-ky-kx."""
        return self._traj_types(self.grid_detection_tolerance)[0]

    @property
    def type_along_k210(self) -> tuple[TrajType, TrajType, TrajType]:
        """Type of trajectory along k2-k1-k0."""
        return self._traj_types(self.grid_detection_tolerance)[1]

    def _traj_types(
        self,
        tolerance: float,
    ) -> tuple[tuple[TrajType, TrajType, TrajType], tuple[TrajType, TrajType, TrajType]]:
        """Calculate the trajectory type along kzkykx and k2k1k0.

        Checks if the entries of the trajectory along certain dimensions
            - are of shape 1 -> `TrajType.SINGLEVALUE`
            - lie on a Cartesian grid -> `TrajType.ONGRID`

        Parameters
        ----------
        tolerance:
            absolute tolerance in checking if points are on integer grid positions

        Returns
        -------
            (`(types along kz,ky,kx)`,`(types along k2,k1,k0)`)

        # TODO: consider non-integer positions that are on a grid, e.g. (0.5, 1, 1.5, ....)
        """
        # Matrix describing trajectory-type [(kz, ky, kx), (k2, k1, k0)]
        # Start with everything not on a grid (arbitrary k-space locations).
        # We use the value of the enum-type to make it easier to do array operations.
        traj_type_matrix = torch.zeros(3, 3, dtype=torch.int)
        for ind, ks in enumerate((self.kz, self.ky, self.kx)):
            values_on_grid = not ks.is_floating_point() or torch.all((ks - ks.round()).abs() <= tolerance)
            for dim in (-3, -2, -1):
                if ks.shape[dim] == 1:
                    traj_type_matrix[ind, dim] |= TrajType.SINGLEVALUE.value | TrajType.ONGRID.value
                if values_on_grid:
                    traj_type_matrix[ind, dim] |= TrajType.ONGRID.value

        # kz should only have flags that are enabled in all columns
        # k2 only flags enabled in all rows, etc
        type_zyx = [TrajType(int(i)) for i in np.bitwise_and.reduce(traj_type_matrix.numpy(), axis=1)]
        type_210 = [TrajType(int(i)) for i in np.bitwise_and.reduce(traj_type_matrix.numpy(), axis=0)]

        # make mypy recognize return  will always have len=3
        return (type_zyx[0], type_zyx[1], type_zyx[2]), (type_210[0], type_210[1], type_210[2])

    def as_tensor(self, stack_dim: int = 0) -> torch.Tensor:
        """Tensor representation of the trajectory.

        Parameters
        ----------
        stack_dim
            The dimension to stack the tensor along.
        """
        shape = self.shape
        return torch.stack([traj.expand(*shape) for traj in (self.kz, self.ky, self.kx)], dim=stack_dim)



"""Indexer class for custom indexing with broadcasting."""

from collections.abc import Sequence
from typing import TypeVar, cast, overload

import torch
import torch.testing
from typing_extensions import Protocol, Self, runtime_checkable

from mrpro.utils.reshape import reduce_view
from mrpro.utils.typing import TorchIndexerType


def normalize_index(ndim: int, index: int) -> int:
    """Normalize possibly negative indices.

    Parameters
    ----------
    ndim
        number of dimensions
    index
        index to normalize. negative indices count from the end.

    Raises
    ------
    `IndexError`
        if index is outside ``[-ndim,ndim)``
    """
    if 0 <= index < ndim:
        return index
    elif -ndim <= index < 0:
        return ndim + index
    else:
        raise IndexError(f'Invalid index {index} for {ndim} data dimensions')


@runtime_checkable
class HasIndex(Protocol):
    """Objects that can be indexed with an `Indexer`."""

    def _index(self, index: 'Indexer') -> Self: ...


T = TypeVar('T', bound=HasIndex)


class Indexer:
    """Custom Indexing with broadcasting.

    This class is used to index tensors in a way that is consistent
    with the shape invariants of the data objects.

    On creation, an index and a shape are required.
    When calling the Indexer with a tensor, the tensor is first broadcasted to the shape,
    then the index is applied to the tensor.

    After indexing, remaining broadcasted dimensions are reduced to singleton dimensions again.
    Thus, using the same Indexer on tensors with different singleton dimensions will
    result in tensors with different shapes. All resulting tensors can be broadcasted
    to the shape that would result in indexing a full tensor already having the desired shape.

    Indexing never removes dimensions, and can only add new dimensions at the beginning of the tensor.

    The index can contain slices, integers, boolean masks, sequences of integers, and integer tensors:
    - Indexing with a slice
        Behaves like in numpy, always returns a view.
        Negative step sizes are not supported and will raise an IndexError.
        slice(None), i.e., means selecting the whole axis.
    - Indexing with an integer
        If the index is within the bounds of the broadcasted shape, indexing behaves like slicing with index:index+1.
        Otherwise, an IndexError is raised.
        Always returns a view.
    - Indexing with a boolean mask
        Singleton dimensions in the mask are interpreted as full slices. This matches broadcasting of the mask to
        the size of the respective axes of the tensor.
        If the mask has more than one non-singleton dimension, a new dimension is added at the beginning of the tensor,
        with length equal to the number of True values in the mask.
        At the indexed axes, singleton dimensions are kept.
        If the mask has only one non-singleton dimension, only the size of the indexed axes is changed.
        Only a single boolean mask is allowed, otherwise an IndexError is raised.
    - Indexing with a sequence of integers
        If a single indexer is a sequence of integers, the result is as if each value of the sequence was used as an
        integer index and the results were concatenated along the indexed dimension.
        If more than one sequence of integers is used, a new dimension at the beginning of the tensor,
        with the length equal to the shape of the sequences, is added. Indexed dimensions are kept as singleton.
        The different sequences must have the same shape, otherwise an IndexError is raised.
        Note that, as in numpy and torch, vectorized indexing is performed, not outer indexing.
        If a single integer tensor is used, the indexed dimension will be replaced by the last dimension of the
        indexing tensor and other dimensions of the indexing tensor are added at the beginning of the result.
    - None
        New axes can be added to the front of tensor by using None in the index.
        This is only allowed at the beginning of the index.
    - Ellipsis
        An indexing expression can contain a single ellipsis, which will be expanded to slice(None)
        for all axes that are not indexed.

    Implementation details:
    - On creation, the indexing expression is parsed and split into two parts: normal_index and fancy_index.
    - normal_index contains only indexing expressions that can be represented as view.
    - fancy_index contains all other indexing expressions.
    - On call
        - the tensor is broadcasted to the desired shape
        - the normal_index is applied.
        - if required, the fancy_index is applied.
        - remaining broadcasted dimensions are reduced to singleton dimensions.
    """

    def __init__(self, broadcast_shape: tuple[int, ...], index: TorchIndexerType) -> None:
        """Initialize the Indexer.

        Parameters
        ----------
        broadcast_shape
            broadcasted shape of the tensor to index. The tensor will be broadcasted to this shape.
        index
            The index to apply to the tensor.
        """
        normal_index: list[slice | int | None] = []
        """Used in phase 1 of the indexing, where we only consider integers and slices. Always does a view"""

        fancy_index: list[slice | torch.Tensor | tuple[int, ...] | None] = []
        """All non normal indices. Might not be possible to do a view."""

        has_fancy_index = False
        """Are there any advanced indices, such as boolean or integer array indices?"""

        vectorized_shape: None | tuple[int, ...] = None
        """Number of dimensions of the integer indices"""

        expanded_index: list[slice | torch.Tensor | tuple[int, ...] | None | int] = []
        """"Index with ellipsis expanded to full slices"""

        # basics checks and figuring out the number of axes already covered by the index,
        # which is needed to determine the number of axes that are covered by the ellipsis
        if not isinstance(index, tuple):
            index = (index,)
        has_ellipsis = False
        has_boolean = False
        covered_axes = 0
        for idx_ in index:
            if idx_ is None:
                if has_ellipsis or covered_axes:
                    raise IndexError('New axes are only allowed at the beginning of the index')
            elif idx_ is Ellipsis:
                if has_ellipsis:
                    raise IndexError('Only one ellipsis is allowed')
                has_ellipsis = True
            elif isinstance(idx_, torch.Tensor) and idx_.dtype == torch.bool:
                if has_boolean:
                    raise IndexError('Only one boolean index is allowed')
                has_boolean = True
                covered_axes += idx_.ndim
            elif isinstance(idx_, int | slice | torch.Tensor) or (
                isinstance(idx_, Sequence) and all(isinstance(el, int) for el in idx_)
            ):
                covered_axes += 1
            else:
                raise IndexError(f'Unsupported index type {idx_}')

        if covered_axes > len(broadcast_shape):
            raise IndexError('Too many indices. Indexing more than the number of available axes is not allowed')

        for idx_ in index:
            if idx_ is Ellipsis:
                # replacing ellipsis with full slices
                expanded_index.extend([slice(None)] * (len(broadcast_shape) - covered_axes))
            elif isinstance(idx_, torch.Tensor | int | slice | None):
                expanded_index.append(idx_)
            else:  # must be Sequence[int], checked above
                # for consistency, we convert all non-tensor sequences of integers to tuples
                expanded_index.append(tuple(cast(Sequence[int], idx_)))

        if not has_ellipsis:
            # if there is no ellipsis, we interpret the index as if it was followed by ellipsis
            expanded_index.extend([slice(None)] * (len(broadcast_shape) - covered_axes))

        number_of_vectorized_indices: int = 0
        shape_position: int = 0  # current position in the shape that we are indexing
        for idx in expanded_index:
            if idx is None:
                # we already checked that None is only allowed at the beginning of the index
                normal_index.append(None)
                fancy_index.append(slice(None))

            elif isinstance(idx, int):
                # always convert integers to slices
                if not -broadcast_shape[shape_position] <= idx < broadcast_shape[shape_position]:
                    raise IndexError(
                        f'Index {idx} out of bounds for axis {shape_position} '
                        f'with shape {broadcast_shape[shape_position]}'
                    )
                normal_index.append(slice(idx, None if idx == -1 else idx + 1))
                fancy_index.append(slice(None))
                shape_position += 1

            elif isinstance(idx, slice):
                if idx.step is not None and idx.step < 0:
                    raise IndexError('Negative step size for slices is not supported')
                normal_index.append(idx)
                fancy_index.append(slice(None))
                shape_position += 1
                continue

            elif isinstance(idx, torch.Tensor) and idx.dtype == torch.bool:
                # boolean indexing
                has_fancy_index = True

                while idx.ndim and idx.shape[0] == 1:
                    # remove leading singleton dimensions and replace by full slices
                    idx = idx.squeeze(0)
                    fancy_index.append(slice(None))
                    normal_index.append(slice(None))
                    shape_position += 1

                right_slice = []
                while idx.ndim and idx.shape[-1] == 1:
                    # remove trailing singleton dimensions and replace by full slices
                    idx = idx.squeeze(-1)
                    right_slice.append(slice(None))

                if idx.ndim == 1:
                    # single boolean dimension remains
                    fancy_index.extend(torch.nonzero(idx, as_tuple=True))
                    number_of_vectorized_indices += 1

                elif idx.ndim > 1:
                    # more than one non singleton dimension
                    for ids, idx_shape, data_shape in zip(
                        idx.nonzero(as_tuple=True),
                        idx.shape,
                        broadcast_shape[shape_position : shape_position + idx.ndim],
                        strict=True,
                    ):
                        if idx_shape == 1:
                            # we interpret singleton dimensions as full slices
                            fancy_index.append(slice(None))
                        elif idx_shape != data_shape:
                            raise IndexError(
                                f'Boolean index has wrong shape, got {idx_shape} but expected {data_shape}'
                            )

                        else:
                            fancy_index.append(ids)
                            number_of_vectorized_indices += 1
                else:
                    # all singleton boolean mask
                    pass

                normal_index.extend(idx.ndim * [slice(None)])
                normal_index.extend(right_slice)
                fancy_index.extend(right_slice)
                shape_position += idx.ndim + len(right_slice)

            elif isinstance(idx, torch.Tensor) and idx.dtype in (
                torch.int64,  # long
                torch.int32,  # int
                torch.int16,
                torch.int8,
                torch.uint16,
                torch.uint32,
                torch.uint64,
            ):
                # integer array indexing
                if (idx >= broadcast_shape[shape_position]).any() or (idx < -broadcast_shape[shape_position]).any():
                    raise IndexError(
                        'Index out of bounds. '
                        f'Got values in the interval [{idx.min()}, {idx.max() + 1}) for axis {shape_position} '
                        f'with shape {broadcast_shape[shape_position]}'
                    )
                if vectorized_shape is not None and vectorized_shape != idx.shape:
                    raise IndexError(
                        f'All vectorized indices must have the same shape. Got {idx.shape} and {vectorized_shape}'
                    )
                vectorized_shape = idx.shape
                has_fancy_index = True
                shape_position += 1
                number_of_vectorized_indices += 1
                normal_index.append(slice(None))
                fancy_index.append(idx.to(torch.int64))

            elif isinstance(idx, tuple):
                # integer Sequence
                if any(el >= broadcast_shape[shape_position] or el < -broadcast_shape[shape_position] for el in idx):
                    raise IndexError(
                        'Index out of bounds. '
                        f'Got values in the interval [{min(idx)}, {max(idx) + 1}) for axis {shape_position} '
                        f'with shape {broadcast_shape[shape_position]}'
                    )
                if vectorized_shape is not None and vectorized_shape != (len(idx),):
                    raise IndexError('All vectorized indices must have the same shape')
                vectorized_shape = (len(idx),)
                has_fancy_index = True
                shape_position += 1
                normal_index.append(slice(None))
                fancy_index.append(idx)
                number_of_vectorized_indices += 1

            else:  # torch.Tensor
                raise IndexError(f'Unsupported index dtype {idx.dtype}')

        self.move_axes: tuple[tuple[int, ...], tuple[int, ...]] = ((), ())
        """final move-axes operation to move the vectorized indices to the beginning of the tensor"""

        self.more_than_one_vectorized_index = number_of_vectorized_indices > 1
        """there is more than one vectorized index, thus a new axis will be added"""

        if self.more_than_one_vectorized_index:
            # torch indexing would remove the dimensions, we want to keep them
            # as singleton dimension -> we need to add a new axis.
            # inserting it in between the indices forces the dimension added by
            # vectorized indices to be always at the beginning of the result.
            self.more_than_one_vectorized_index = True
            new_fancy_index = []
            for idx in fancy_index:
                new_fancy_index.append(idx)
                if isinstance(idx, torch.Tensor):
                    new_fancy_index.append(None)
                if isinstance(idx, tuple):
                    new_fancy_index.append(None)
            fancy_index = new_fancy_index

        elif vectorized_shape is not None and len(vectorized_shape) != 1:
            # for a single vectorized index, torch would insert it at the same position
            # this would shift the other axes, potentially causing violations of the shape invariants.
            # thus, we move the inserted axis to the beginning of the tensor, after axes inserted by None.
            # We keep the last axes of the vectorized index in the indexed axis.
            move_source_start = next(i for i, idx in enumerate(fancy_index) if isinstance(idx, torch.Tensor))
            move_source = tuple(range(move_source_start, move_source_start + len(vectorized_shape) - 1))
            move_target_start = next(i for i, idx in enumerate(fancy_index) if idx is not None)
            move_target = tuple(range(move_target_start, move_target_start + len(vectorized_shape) - 1))
            self.move_axes = (move_source, move_target)

        self.fancy_index = tuple(fancy_index) if has_fancy_index else ()
        self.normal_index = tuple(normal_index)
        self.shape = broadcast_shape

    @overload
    def __call__(self, tensor: torch.Tensor) -> torch.Tensor: ...
    @overload
    def __call__(self, tensor: T) -> T: ...
    def __call__(self, tensor: torch.Tensor | T) -> torch.Tensor | T:
        """Apply the index to a tensor or object implementing _index."""
        if isinstance(tensor, HasIndex):
            return cast(T, tensor._index(self))
        try:
            tensor = tensor.broadcast_to(self.shape)
        except RuntimeError:
            raise IndexError('Tensor cannot be broadcasted to the desired shape') from None

        tensor = tensor[self.normal_index]  # will always be a view

        if not self.fancy_index:
            # nothing more to do
            tensor = reduce_view(tensor)
            return tensor

        # we need to modify the fancy index to efficiently handle broadcasted dimensions
        fancy_index: list[None | tuple[int, ...] | torch.Tensor | slice] = []
        tensor_index = 0
        stride = tensor.stride()
        for idx in self.fancy_index:
            if idx is None:
                fancy_index.append(idx)
                # don't increment tensor_index as this is a new axis
                continue
            if stride[tensor_index] == 0:
                # broadcasted dimension
                if isinstance(idx, slice):  # can only be slice(None) here
                    # collapse broadcasted dimensions to singleton, i.e. keep the dimension
                    fancy_index.append(slice(0, 1))
                elif not self.more_than_one_vectorized_index and isinstance(idx, tuple):
                    # as the dimension only exists due to broadcasting, it should be reduced to singleton
                    # there is already a None inserted after the index, so we don't need to keep the dimension
                    fancy_index.append((0,))
                elif not self.more_than_one_vectorized_index and isinstance(idx, torch.Tensor):
                    # same, but with more dimensions in the single vectorized index
                    # these axes will later be moved to the beginning of the tensor, as they would
                    # if the dimensions were not broadcasted
                    fancy_index.append(idx.new_zeros([1] * idx.ndim))
                else:
                    fancy_index.append(idx)
            else:
                fancy_index.append(idx)
            tensor_index += 1

        tensor = tensor[tuple(fancy_index)]

        if self.move_axes[0]:
            # handle the special case of a single and integer index, where we need to move the new
            # axis to the beginning of the tensor
            tensor = tensor.moveaxis(self.move_axes[0], self.move_axes[1])

        return tensor



"""MD-CNN radial cardiac cine dataset."""

import concurrent.futures
import io
import re
import urllib.request
from collections.abc import Sequence
from os import PathLike
from pathlib import Path

import numpy as np
import platformdirs
import scipy.io
import torch
from einops import rearrange
from torch.utils.data import Dataset
from tqdm import tqdm

from mrpro.data.AcqInfo import AcqInfo
from mrpro.data.enums import TrajectoryType
from mrpro.data.KData import KData
from mrpro.data.KHeader import KHeader
from mrpro.data.SpatialDimension import SpatialDimension
from mrpro.data.traj_calculators.KTrajectoryRadial2D import KTrajectoryRadial2D

# ~/.cache/mrpro/mdcnn on Linux, %AppData%\Local\mrpro\mdcnn on Windows
CACHE_DIR_MDCNN = Path(platformdirs.user_cache_dir('mrpro')) / 'mdcnn'


def download_mdcnn(
    output_directory: str | PathLike = CACHE_DIR_MDCNN,
    n_files: int = 108,
    workers: int = 4,
    progress: bool = False,
) -> None:
    """
    Download MD-CNN [MDCNN]_ dataset from Harvard Dataverse and save it as numpy files.

    References
    ----------
    .. [MDCNN] H. El-Rewaidy, “Replication Data for: Multi-Domain Convolutional Neural Network (MD-CNN) For Radial
       Reconstruction of Dynamic Cardiac MRI.” Harvard Dataverse, 2020. https://doi.org/10.7910/DVN/CI3WB6.

    Parameters
    ----------
    output_directory : Path
        Directory to save files.
    n_files : int
        Number of files to download. Maximum is 108.
        If lower than 108, the first `n_files` based on the harvard dataverse id will be downloaded.
    workers : int
        Number of parallel downloads.
    progress : bool
        Show progress bar.
    """
    progressbar = tqdm(total=n_files, desc='Downloading') if progress else None
    if not 0 < n_files <= 108:
        raise ValueError('n_files must be positive and less than 108')
    try:
        output_directory_ = Path(output_directory)
        output_directory_.mkdir(parents=True, exist_ok=True)
    except Exception as e:  # noqa: BLE001
        raise RuntimeError(f'Failed to create directory {output_directory}: {e}') from None

    def fetch(file_id: int) -> None:
        nonlocal n_files
        url = f'https://dataverse.harvard.edu/api/access/datafile/{file_id}'
        with urllib.request.urlopen(urllib.request.Request(url, headers={'User-Agent': 'mrpro'})) as resp:  # noqa: S310
            if (match := re.search('(P[0-9]+)_', resp.headers['Content-Disposition'])) is None:
                return
            if n_files < 1:
                return
            n_files -= 1
            subject = match.groups()[0]
            data = torch.as_tensor(scipy.io.loadmat(io.BytesIO(resp.read()))['data'], dtype=torch.float32)
            data = torch.view_as_complex(
                rearrange(data, '1 phase k0 k1 coils complex -> phase coils 1 k1 k0 complex').contiguous()
            )
            if data[0, 0, 0, 0, 208].abs() > 1e-12:
                first = 208
            else:
                first = 296
            data = data[..., first:624]
            # we save as numpy, as loadmat is more than 10x slower than np.load
            filename = output_directory_ / f'{subject}.npy'
            np.save(filename, data)
            if progressbar is not None:
                progressbar.update(1)

    with concurrent.futures.ThreadPoolExecutor(max_workers=workers) as executor:
        for file_id in range(4000844, 4000978):
            executor.submit(fetch, file_id)


class MDCNNDataset(Dataset):
    """MD-CNN radial cardiac cine dataset.

     The MD-CNN [MDCNN]_ dataset is a radial cardiac cine dataset with 108 subjects.
     See `mrpro.phantoms.mdcnn.download_mdcnn` for downloading and conversion to numpy format of the dataset.

     The dataset returns `mrpro.data.KData` objects.

    References
    ----------
    .. [MDCNN] El-Rewaidy H., Replication Data for: Multi-Domain Convolutional Neural Network (MD-CNN) For Radial
       Reconstruction of Dynamic Cardiac MRI. Harvard Dataverse, 2020. https://doi.org/10.7910/DVN/CI3WB6.
    .. [REWAIDY] El-Rewaidy H, Fahmy AS, Pashakhanloo F, et al. Multi-domain convolutional neural network (MD-CNN) for
       radial reconstruction of dynamic cardiac MRI. Magn Reson Med. 2020; 85: 1195-1208. https://doi.org/10.1002/mrm.28485
    """

    def __init__(self, path: str | PathLike | Sequence[str | PathLike] = CACHE_DIR_MDCNN):
        """Initialize the MD-CNN dataset.

        Parameters
        ----------
        path : str | PathLike
            Path to the directory containing the MD-CNN dataset converted to numpy format or sequence of npy files.
        """
        files = list(Path(path).rglob('P*.npy')) if isinstance(path, str | PathLike) else [Path(p) for p in path]
        self.files = sorted(files, key=lambda x: int(x.stem[1:]))

    def __len__(self) -> int:
        """Get the number of CINES in the dataset."""
        return len(self.files)

    def __getitem__(self, idx: int) -> KData:
        """Get one CINE from the dataset."""
        filename = self.files[idx]
        data = torch.as_tensor(np.load(filename))
        n_phases, *_, n_k1, n_k0 = data.shape

        if n_k0 == 416:  # different partial fourier settings
            k0_center = 417 - 208
        else:
            k0_center = 417 - 296

        info = AcqInfo()
        info.idx.k1 = torch.arange(n_k1)[None, None, None, :, None] + n_k1 // 2
        info.idx.phase = torch.arange(n_phases)[:, None, None, None, None]
        traj = KTrajectoryRadial2D(torch.pi / n_k1)(
            n_k0=n_k0,
            k0_center=k0_center,
            k1_idx=info.idx.k1,
        )

        header = KHeader(
            encoding_matrix=SpatialDimension(1, 624, 624),
            recon_matrix=SpatialDimension(1, 208, 208),
            acq_info=info,
            recon_fov=SpatialDimension(1, 0.38, 0.38),
            encoding_fov=SpatialDimension(1, 1.52, 1.52),
            vendor='Siemens',
            model='MAGNETOM Vida',
            protocol_name='bSSFP Cine',
            tr=torch.tensor(3.06e-3)[None, None, None, None, None],
            te=torch.tensor(1.4e-3)[None, None, None, None, None],
            fa=torch.deg2rad(torch.tensor(48))[None, None, None, None, None],
            lamor_frequency_proton=123e6,
            trajectory_type=TrajectoryType.RADIAL,
            patient_name=filename.stem,
        )
        return KData(data=data, header=header, traj=traj)



"""Cardiac MR Fingerprinting signal model (EPG)."""

from collections.abc import Sequence

import torch

from mrpro.operators.models.EPG import DelayBlock, EPGSequence, FispBlock, InversionBlock, Parameters, T2PrepBlock
from mrpro.operators.SignalModel import SignalModel


class CardiacFingerprinting(SignalModel[torch.Tensor, torch.Tensor, torch.Tensor]):
    """Cardiac MR Fingerprinting signal model.

    This model simulates a cardiac MR fingerprinting sequence as described in [HAMI2017]_ and [HAMI2020]_ using the
    extended phase graph (`~mrpro.operators.models.EPG`) formalism.

    It is a three-fold repetition of

    .. code-block:: text

        Block 0          Block 1          Block 2          Block 3          Block 4
        R-peak           R-peak           R-peak           R-peak           R-peak
        |----------------|----------------|----------------|----------------|----------------
         [INV 30ms][ACQ]           [ACQ]   [T2-prep][ACQ]   [T2-prep][ACQ]  [T2-prep][ACQ]

    Note
    ----
    This model is on purpose not flexible in all design choices. Instead, consider writing a custom
    `~mrpro.operators.SignalModel` based on this implementation if you need to simulate a different sequence.

    References
    ----------
    .. [HAMI2017] Hamilton, J. I. et al. (2017) MR fingerprinting for rapid quantification of myocardial T1, T2, and
           proton spin density. Magn. Reson. Med. 77 http://doi.wiley.com/10.1002/mrm.26216
    .. [HAMI2020] Hamilton, J.I. et al. (2020) Simultaneous Mapping of T1 and T2 Using Cardiac Magnetic Resonance
           Fingerprinting in a Cohort of Healthy Subjects at 1.5T. J Magn Reson Imaging, 52: 1044-1052. https://doi.org/10.1002/jmri.27155
    """

    def __init__(
        self,
        acquisition_times: torch.Tensor | Sequence[int] = (0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14),
        echo_time: float = 0.0015,
        repetition_time: float = 0.005,
        t2_prep_echo_times: tuple[float, float, float] = (0.03, 0.05, 0.08),
    ) -> None:
        """Initialize the Cardiac MR Fingerprinting signal model.

        Parameters
        ----------
        acquisition_times
            Times of the acquisistions in s.
            Either 15 values (the first acquisition of each block) or 705 values for all acquisitions.
            In both cases only the time of the first acquisition of each block will be used to calculate the delays.

        echo_time
            TE in s
        repetition_time
            TR in s
        t2_prep_echo_times
            Echo times of the three T2 preparation blocks in s
        """
        super().__init__()
        sequence = EPGSequence()
        max_flip_angles_deg = [12.5, 18.75, 25, 25, 25, 12.5, 18.75, 25, 25.0, 25, 12.5, 18.75, 25, 25, 25]
        if len(acquisition_times) == 15:
            block_time = torch.as_tensor(acquisition_times)
        elif len(acquisition_times) == 705:
            block_time = torch.as_tensor(acquisition_times[::47])
        else:
            raise ValueError(f'Invalid number of acquisition times: {len(acquisition_times)}. Must be 15 or 705.')
        for i in range(15):
            block = EPGSequence()
            match i % 5:
                case 0:
                    block.append(InversionBlock(0.020))
                case 1:
                    pass
                case 2:
                    block.append(T2PrepBlock(t2_prep_echo_times[0]))
                case 3:
                    block.append(T2PrepBlock(t2_prep_echo_times[1]))
                case 4:
                    block.append(T2PrepBlock(t2_prep_echo_times[2]))
            flip_angles = torch.deg2rad(
                torch.cat((torch.linspace(4, max_flip_angles_deg[i], 16), torch.full((31,), max_flip_angles_deg[i])))
            )
            block.append(FispBlock(flip_angles, 0.0, tr=repetition_time, te=echo_time))

            if i > 0:
                delay = (block_time[i] - block_time[i - 1]) - block.duration
                if (delay < 0).any():
                    raise ValueError(f'Block {i} would start before the previous block finished. ')
                sequence.append(DelayBlock(delay))
            sequence.append(block)
        self.sequence = sequence.to(device=block_time.device)

    def __call__(self, m0: torch.Tensor, t1: torch.Tensor, t2: torch.Tensor) -> tuple[torch.Tensor]:
        """Simulate the Cardiac MR Fingerprinting (cMRF) signal.

        Parameters
        ----------
        m0
            Equilibrium signal / proton density. (complex).
            Shape `(...)`, for example `(*other, coils, z, y, x)` or `(samples)`.
        t1
            Longitudinal (T1) relaxation time in seconds.
            Shape `(...)`, for example `(*other, coils, z, y, x)` or `(samples)`.
        t2
            Transversal (T2) relaxation time in seconds.
            Shape `(...)`, for example `(*other, coils, z, y, x)` or `(samples)`.

        Returns
        -------
            Simulated Cardiac MR Fingerprinting signal.
            Shape `(acquisitions ...)`, for example `(acquisitions, *other, coils, z, y, x)` or
            `(acquisitions, samples)` where `acquisitions` corresponds to the different acquisitions
            in the sequence.
        """
        return super().__call__(m0, t1, t2)

    def forward(self, m0: torch.Tensor, t1: torch.Tensor, t2: torch.Tensor) -> tuple[torch.Tensor]:
        """Apply forward of CardiacFingerprinting.

        .. note::
            Prefer calling the instance of the CardiacFingerprinting as ``operator(x)`` over
            directly calling this method. See this PyTorch `discussion <https://discuss.pytorch.org/t/is-model-forward-x-the-same-as-model-call-x/33460/3>`_.
        """
        parameters = Parameters(m0, t1, t2)
        _, signals = self.sequence(parameters, states=20)
        signal = torch.stack(signals, dim=0)
        return (signal,)



"""L1 Norm with :math:`C` as :math:`R^2`."""

import torch

from mrpro.operators.Functional import ElementaryProximableFunctional, throw_if_negative_or_complex


class L1NormViewAsReal(ElementaryProximableFunctional):
    r"""Functional class for the L1 Norm, where C is identified with :math:`R^2`.

    This implements the functional given by
    :math:`f: C^N \rightarrow [0, \infty), x \rightarrow \|W_r * \mathrm{Re}(x-b))\|_1 +\|( W_i *\mathrm{Im}(x-b))\|_1`,
    where :math:`W_r` and :math:`W_i` are a either scalars or tensors and `*` denotes element-wise multiplication.

    If the parameter `weight` is real-valued, :math:`W_r` and :math:`W_i` are both set to `weight`.
    If it is complex-valued, :math:`W_r` and :math:`W_I` are set to the real and imaginary part, respectively.

    In most cases, consider setting `divide_by_n` to `true` to be independent of input size.

    The norm of the vector is computed along the dimensions set at initialization.
    """

    def __call__(
        self,
        x: torch.Tensor,
    ) -> tuple[torch.Tensor]:
        r"""Compute the L1 norm, viewing complex numbers as R^2.

        Calculates :math:`\|W_r * \mathrm{Re}(x-b)\|_1 + \|W_i * \mathrm{Im}(x-b)\|_1`.
        If `weight` is real, :math:`W_r = W_i = \mathrm{weight}`.
        If `weight` is complex, :math:`W_r = \mathrm{Re}(\mathrm{weight})` and :math:`W_i=\mathrm{Im}(\mathrm{weight})`.
        `b` is `target`. The norm is computed along `dim`.
        If `divide_by_n` is `True`, the result is averaged; otherwise, summed.

        Parameters
        ----------
        x
            Input tensor.

        Returns
        -------
            The L1 norm. If `keepdim` is `True`, the dimensions `dim` are retained
            with size 1; otherwise, they are reduced.
        """
        return super().__call__(x)

    def forward(
        self,
        x: torch.Tensor,
    ) -> tuple[torch.Tensor]:
        """Apply forward of L1NormViewAsReal.

        .. note::
            Prefer calling the instance of the L1NormViewAsReal operator as ``operator(x)`` over
            directly calling this method. See this PyTorch `discussion <https://discuss.pytorch.org/t/is-model-forward-x-the-same-as-model-call-x/33460/3>`_.
        """
        dtype = torch.promote_types(self.target.dtype, x.dtype)
        x = x.to(dtype)
        target = self.target.to(dtype)
        diff = x - target
        if diff.is_complex() and self.weight.is_complex():
            value = (self.weight.real * diff.real).abs() + (self.weight.imag * diff.imag).abs()
        elif diff.is_complex():
            value = (self.weight * diff.real).abs() + (self.weight * diff.imag).abs()
        else:
            value = (self.weight * diff).abs()

        if self.divide_by_n:
            return (torch.mean(value, dim=self.dim, keepdim=self.keepdim),)
        else:
            return (torch.sum(value, dim=self.dim, keepdim=self.keepdim),)

    def prox(self, x: torch.Tensor, sigma: torch.Tensor | float = 1.0) -> tuple[torch.Tensor]:
        """Proximal Mapping of the L1 Norm.

        Apply the proximal mapping of the L1 norm with :math:`C` identified as :math:`R^2`.

        Parameters
        ----------
        x
            input tensor
        sigma
            real valued positive scaling factor

        Returns
        -------
            Proximal mapping applied to the input tensor
        """
        throw_if_negative_or_complex(sigma)
        diff = x - self.target
        threshold = self._divide_by_n(self.weight * sigma, torch.broadcast_shapes(x.shape, self.weight.shape))
        out = torch.sgn(diff.real) * torch.relu(diff.real.abs() - threshold.real.abs())
        if diff.is_complex():
            threshold_imag = threshold.imag if self.weight.is_complex() else threshold
            imag = torch.sgn(diff.imag) * torch.relu(diff.imag.abs() - threshold_imag.abs())
            out = torch.complex(out, imag)
        out = out + self.target
        out = out.to(torch.result_type(threshold, out))
        return (out,)



"""Data containers, loading and saving data."""

from mrpro.data import enums, traj_calculators, acq_filters
from mrpro.data.AcqInfo import AcqIdx, AcqInfo
from mrpro.data.CsmData import CsmData
from mrpro.data.Dataclass import Dataclass, InconsistentDeviceError
from mrpro.data.DcfData import DcfData
from mrpro.data.EncodingLimits import EncodingLimits, Limits
from mrpro.data.IData import IData
from mrpro.data.IHeader import IHeader
from mrpro.data.KData import KData
from mrpro.data.KHeader import KHeader
from mrpro.data.KNoise import KNoise
from mrpro.data.KTrajectory import KTrajectory
from mrpro.data.QData import QData
from mrpro.data.QHeader import QHeader
from mrpro.data.Rotation import Rotation
from mrpro.data.SpatialDimension import SpatialDimension

__all__ = [
    "AcqIdx",
    "AcqInfo",
    "CsmData",
    "Dataclass",
    "DcfData",
    "EncodingLimits",
    "IData",
    "IHeader",
    "InconsistentDeviceError",
    "KData",
    "KHeader",
    "KNoise",
    "KTrajectory",
    "Limits",
    "QData",
    "QHeader",
    "Rotation",
    "SpatialDimension",
    "acq_filters",
    "enums",
    "traj_calculators"
]



"""MR quantitative data header (QHeader) dataclass."""

from dataclasses import field

import torch
from pydicom.dataset import Dataset
from pydicom.tag import Tag
from typing_extensions import Self

from mrpro.data.Dataclass import Dataclass
from mrpro.data.IHeader import IHeader, ImageIdx
from mrpro.data.KHeader import KHeader
from mrpro.data.Rotation import Rotation
from mrpro.data.SpatialDimension import SpatialDimension
from mrpro.utils.unit_conversion import mm_to_m


class QHeader(Dataclass):
    """MR quantitative data header."""

    resolution: SpatialDimension[float]
    """Resolution [m/px] of the data."""

    position: SpatialDimension[torch.Tensor] = field(
        default_factory=lambda: SpatialDimension(
            torch.zeros(1, 1, 1, 1, 1),
            torch.zeros(1, 1, 1, 1, 1),
            torch.zeros(1, 1, 1, 1, 1),
        )
    )
    """Position of the data"""

    orientation: Rotation = field(default_factory=lambda: Rotation.identity((1, 1, 1, 1, 1)))
    """Orientation of the data"""

    idx: ImageIdx = field(default_factory=ImageIdx)

    @classmethod
    def from_iheader(cls, header: IHeader) -> Self:
        """Create QHeader object from IHeader object.

        Parameters
        ----------
        header
            MR raw data header (IHeader) containing required meta data.
        """
        return cls(resolution=header.resolution, position=header.position, orientation=header.orientation)

    @classmethod
    def from_kheader(cls, header: KHeader) -> Self:
        """Create QHeader object from KHeader object.

        Parameters
        ----------
        header
            MR raw data header (KHeader) containing required meta data.
        """
        resolution = header.recon_fov / header.recon_matrix
        return cls(resolution=resolution, position=header.acq_info.position, orientation=header.acq_info.orientation)

    @classmethod
    def from_dicom(cls, dicom_dataset: Dataset) -> Self:
        """Read DICOM file containing qMRI data and return QHeader object.

        Parameters
        ----------
        dicom_dataset
            dataset object containing the DICOM file.
        """

        def get_items(name: str):  # Todo: move to utils and reuse logic in IHeader
            """Get all items with a given name from a pydicom dataset."""
            # iterall is recursive, so it will find all items with the given name
            return [item.value for item in dicom_dataset.iterall() if item.tag == Tag(name)]

        resolution = SpatialDimension(
            get_items('SliceThickness')[0], get_items('PixelSpacing')[0][1], get_items('PixelSpacing')[0][0]
        ).apply_(mm_to_m)
        return cls(resolution=resolution)



"""Conversion between different units."""

from typing import TypeVar

import numpy as np
import torch

__all__ = [
    'GYROMAGNETIC_RATIO_PROTON',
    'deg_to_rad',
    'lamor_frequency_to_magnetic_field',
    'm_to_micrometer',
    'm_to_mm',
    'magnetic_field_to_lamor_frequency',
    'micrometer_to_m',
    'mm_to_m',
    'ms_to_s',
    'rad_to_deg',
    's_to_ms',
]

GYROMAGNETIC_RATIO_PROTON = 42.5764 * 1e6
r"""The gyromagnetic ratio :math:`\frac{\gamma}{2\pi}` of 1H in H2O in Hz/T."""

# Conversion functions for units
T = TypeVar('T', float, torch.Tensor, list[float], tuple[float, ...])


def ms_to_s(ms: T) -> T:
    """Convert ms to s."""
    if isinstance(ms, list):
        return [ms_to_s(x) for x in ms]
    if isinstance(ms, tuple):
        return tuple([ms_to_s(x) for x in ms])
    """Convert ms to s."""
    return ms / 1000


def s_to_ms(s: T) -> T:
    """Convert s to ms."""
    if isinstance(s, list):
        return [s_to_ms(x) for x in s]
    if isinstance(s, tuple):
        return tuple([s_to_ms(x) for x in s])
    return s * 1000


def mm_to_m(mm: T) -> T:
    """Convert mm to m."""
    if isinstance(mm, list):
        return [mm_to_m(x) for x in mm]
    if isinstance(mm, tuple):
        return tuple([mm_to_m(x) for x in mm])
    return mm / 1000


def m_to_mm(m: T) -> T:
    """Convert m to mm."""
    if isinstance(m, list):
        return [m_to_mm(x) for x in m]
    if isinstance(m, tuple):
        return tuple([m_to_mm(x) for x in m])
    return m * 1000


def micrometer_to_m(micrometer: T) -> T:
    """Convert micrometer to m."""
    if isinstance(micrometer, list):
        return [micrometer_to_m(x) for x in micrometer]
    if isinstance(micrometer, tuple):
        return tuple([micrometer_to_m(x) for x in micrometer])
    return micrometer / 1e6


def m_to_micrometer(m: T) -> T:
    """Convert m to micrometer."""
    if isinstance(m, list):
        return [m_to_micrometer(x) for x in m]
    if isinstance(m, tuple):
        return tuple([m_to_micrometer(x) for x in m])
    return m * 1e6


def deg_to_rad(deg: T) -> T:
    """Convert degree to radians."""
    if isinstance(deg, torch.Tensor):
        return torch.deg2rad(deg)
    if isinstance(deg, tuple):
        return tuple([deg_to_rad(x) for x in deg])
    if isinstance(deg, list):
        return [deg_to_rad(x) for x in deg]
    return deg / 180.0 * np.pi


def rad_to_deg(rad: T) -> T:
    """Convert radians to degree."""
    if isinstance(rad, torch.Tensor):
        return torch.rad2deg(rad)
    if isinstance(rad, tuple):
        return tuple([rad_to_deg(x) for x in rad])
    if isinstance(rad, list):
        return [rad_to_deg(x) for x in rad]
    return rad * 180.0 / np.pi


def lamor_frequency_to_magnetic_field(lamor_frequency: T, gyromagnetic_ratio: float = GYROMAGNETIC_RATIO_PROTON) -> T:
    """Convert the Lamor frequency [Hz] to the magntic field strength [T].

    Parameters
    ----------
    lamor_frequency
        Lamor frequency [Hz]
    gyromagnetic_ratio
        Gyromagnetic ratio [Hz/T], default: gyromagnetic ratio of 1H proton

    Returns
    -------
        Magnetic field strength [T]
    """
    if isinstance(lamor_frequency, list):
        return [lamor_frequency_to_magnetic_field(x) for x in lamor_frequency]
    if isinstance(lamor_frequency, tuple):
        return tuple([lamor_frequency_to_magnetic_field(x) for x in lamor_frequency])
    return lamor_frequency / gyromagnetic_ratio


def magnetic_field_to_lamor_frequency(
    magnetic_field_strength: T, gyromagnetic_ratio: float = GYROMAGNETIC_RATIO_PROTON
) -> T:
    """Convert the magntic field strength [T] to Lamor frequency [Hz].

    Parameters
    ----------
    magnetic_field_strength
       Strength of the magnetic field [T]
    gyromagnetic_ratio
        Gyromagnetic ratio [Hz/T], default: gyromagnetic ratio of 1H proton

    Returns
    -------
        Lamor frequency [Hz]
    """
    if isinstance(magnetic_field_strength, tuple):
        return tuple([magnetic_field_to_lamor_frequency(x) for x in magnetic_field_strength])
    if isinstance(magnetic_field_strength, list):
        return [magnetic_field_to_lamor_frequency(x) for x in magnetic_field_strength]
    return magnetic_field_strength * gyromagnetic_ratio



"""Total Variation (TV)-Regularized Reconstruction using PDHG."""

from __future__ import annotations

from collections.abc import Callable, Sequence

import torch

from mrpro.algorithms.optimizers.pdhg import pdhg
from mrpro.algorithms.prewhiten_kspace import prewhiten_kspace
from mrpro.algorithms.reconstruction.DirectReconstruction import DirectReconstruction
from mrpro.data.CsmData import CsmData
from mrpro.data.DcfData import DcfData
from mrpro.data.IData import IData
from mrpro.data.KData import KData
from mrpro.data.KNoise import KNoise
from mrpro.operators import LinearOperatorMatrix, ProximableFunctionalSeparableSum
from mrpro.operators.FiniteDifferenceOp import FiniteDifferenceOp
from mrpro.operators.functionals import L1NormViewAsReal, L2NormSquared
from mrpro.operators.LinearOperator import LinearOperator
from mrpro.utils import normalize_index, unsqueeze_right


class TotalVariationRegularizedReconstruction(DirectReconstruction):
    r"""TV-regularized reconstruction.

    This algorithm solves the problem :math:`min_x \frac{1}{2}||Ax - y||_2^2 + \sum_i l_i ||\nabla_i x||_1`
    by using the PDHG-algorithm. :math:`A` is the acquisition model (coil sensitivity maps, Fourier operator,
    k-space sampling), :math:`y` is the acquired k-space data, :math:`l_i` are the strengths of the regularization
    along the different dimensions and :math:`\nabla_i` is the finite difference operator applied to :math:`x` along
    different dimensions :math:`i`.
    """

    max_iterations: int
    """Maximum number of PDHG iterations."""

    tolerance: float
    """Tolerance of PDHG for relative change of the primal solution."""

    regularization_dim: Sequence[int]
    """Dimensions along which the total variation reguarization is applied :math:`i`."""

    regularization_weight: torch.Tensor
    """Strengths of the regularization along different dimensions :math:`l_i`."""

    def __init__(
        self,
        kdata: KData | None = None,
        fourier_op: LinearOperator | None = None,
        csm: Callable | CsmData | None = CsmData.from_idata_walsh,
        noise: KNoise | None = None,
        dcf: DcfData | None = None,
        *,
        max_iterations: int = 100,
        tolerance: float = 0,
        regularization_dim: Sequence[int],
        regularization_weight: float | Sequence[float] | Sequence[torch.Tensor],
    ) -> None:
        """Initialize TotalVariationRegularizedReconstruction.

        Parameters
        ----------
        kdata
            KData. If `kdata` is provided and `fourier_op` or `dcf` are `None`, then `fourier_op` and `dcf` are
            estimated based on `kdata`. Otherwise `fourier_op` and `dcf` are used as provided.
        fourier_op
            Instance of the `~mrpro.operators.FourierOp` used for reconstruction. If `None`, set up based on `kdata`.
        csm
            Sensitivity maps for coil combination. If `None`, no coil combination is carried out, i.e. images for each
            coil are returned. If a `Callable` is provided, coil images are reconstructed using the adjoint of the
            `~mrpro.operators.FourierOp` (including density compensation) and then sensitivity maps are calculated
            using the `Callable`. For this, `kdata` needs also to be provided.
            For examples have a look at the `mrpro.data.CsmData` class e.g. `~mrpro.data.CsmData.from_idata_walsh`
            or `~mrpro.data.CsmData.from_idata_inati`.
        noise
            KNoise used for prewhitening. If `None`, no prewhitening is performed
        dcf
            K-space sampling density compensation. If `None`, set up based on `kdata`. The `dcf` is only used to
            calculate a starting estimate for PDHG.
        max_iterations
            Maximum number of PDHG iterations
        tolerance
            Tolerance of PDHG for relative change of the primal solution; if zero, `max_iterations` of PDHG are run.
        regularization_dim
            Dimensions along which the total variation reguarization is applied (:math:`i`).
        regularization_weight
            Strengths of the regularization (:math:`l_i`). If a single values is given, it is applied to all dimensions.
            If a sequence is given, it must have the same length as `regularization_dim`.

        Raises
        ------
        ValueError
            If the `kdata` and `fourier_op` are `None` or if `csm` is a `Callable` but `kdata` is `None`.
        ValueError
            If `regularization_dim` contains repeated values.
        ValueError
            If the length of `regularization_dim` and `regularization_weight` do not match
        """
        super().__init__(kdata, fourier_op, csm, noise, dcf)
        self.max_iterations = max_iterations
        self.tolerance = tolerance

        if len(regularization_dim) != len(set(regularization_dim)):
            raise ValueError('Repeated values are not allowed in regularization_dim')
        self.regularization_dim = regularization_dim

        if isinstance(regularization_weight, float):
            regularization_weight = [regularization_weight] * len(regularization_dim)
        if len(regularization_dim) != len(regularization_weight):
            raise ValueError('Regularization dimensions and weights must have the same length')
        self.regularization_weight = torch.as_tensor(regularization_weight)

    def forward(self, kdata: KData) -> IData:
        """Apply the reconstruction.

        Parameters
        ----------
        kdata
            k-space data to reconstruct.

        Returns
        -------
            the reconstruced image.
        """
        regularization_dim = tuple(normalize_index(kdata.ndim, idx) for idx in self.regularization_dim)
        if len(regularization_dim) != len(set(regularization_dim)):
            raise ValueError('Repeated values are not allowed in regularization_dim')

        if self.noise is not None:
            kdata = prewhiten_kspace(kdata, self.noise)

        acquisition_operator = self.fourier_op @ self.csm.as_operator() if self.csm is not None else self.fourier_op
        l2_norm_squared = L2NormSquared(target=kdata.data)

        # TV regularization
        nabla_operator = FiniteDifferenceOp(dim=regularization_dim, mode='forward')
        l1_norm = L1NormViewAsReal(weight=unsqueeze_right(self.regularization_weight, kdata.data.ndim))
        operator = LinearOperatorMatrix(((acquisition_operator,), (nabla_operator,)))

        initial_value = acquisition_operator.H(
            self.dcf.as_operator()(kdata.data)[0] if self.dcf is not None else kdata.data
        )[0]

        (img_tensor,) = pdhg(
            f=ProximableFunctionalSeparableSum(l2_norm_squared, l1_norm),
            g=None,
            operator=operator,
            initial_values=(initial_value,),
            max_iterations=self.max_iterations,
            tolerance=self.tolerance,
        )
        img = IData.from_tensor_and_kheader(img_tensor, kdata.header)
        return img



"""Encoding limits dataclass."""

import dataclasses
from dataclasses import field

from ismrmrd.xsd.ismrmrdschema.ismrmrd import encodingLimitsType, encodingType, ismrmrdHeader, limitType
from typing_extensions import Self


@dataclasses.dataclass(slots=True)
class Limits:
    """Limits dataclass with min, max, and center attributes."""

    min: int = 0
    """Lower boundary."""

    max: int = 0
    """Upper boundary."""

    center: int = 0
    """Center."""

    @classmethod
    def from_ismrmrd(cls, limit_type: limitType | None) -> Self:
        """Create Limits from ismrmrd.limitType."""
        if limit_type is None:
            return cls()
        return cls(*dataclasses.astuple(limit_type))

    def to_ismrmrd(self) -> limitType:
        """Convert Limits to ismsmrd.limitType."""
        return limitType(self.min, self.max, self.center)

    @property
    def length(self) -> int:
        """Length of the limits."""
        return self.max - self.min + 1


@dataclasses.dataclass(slots=True)
class EncodingLimits:
    """Encoding limits dataclass with limits for each attribute [INA2016]_.

    References
    ----------
    .. [INA2016] Inati S, Hansen M (2016) ISMRM Raw data format: A proposed standard for MRI raw datasets. MRM 77(1)
        https://doi.org/10.1002/mrm.26089

    """

    k0: Limits = field(default_factory=Limits)
    """First k-space encoding."""

    k1: Limits = field(default_factory=Limits)
    """Second k-space encoding."""

    k2: Limits = field(default_factory=Limits)
    """Third k-space encoding."""

    average: Limits = field(default_factory=Limits)
    """Signal average."""

    slice: Limits = field(default_factory=Limits)
    """Slice number (multi-slice 2D)."""

    contrast: Limits = field(default_factory=Limits)
    """Echo number in multi-echo."""

    phase: Limits = field(default_factory=Limits)
    """Cardiac phase."""

    repetition: Limits = field(default_factory=Limits)
    """Repeated/dynamic acquisitions."""

    set: Limits = field(default_factory=Limits)
    """Sets of different preparation."""

    segment: Limits = field(default_factory=Limits)
    """Segments of segmented acquisition."""

    user0: Limits = field(default_factory=Limits)
    """User index 0."""

    user1: Limits = field(default_factory=Limits)
    """User index 1."""

    user2: Limits = field(default_factory=Limits)
    """User index 2."""

    user3: Limits = field(default_factory=Limits)
    """User index 3."""

    user4: Limits = field(default_factory=Limits)
    """User index 4."""

    user5: Limits = field(default_factory=Limits)
    """User index 5."""

    user6: Limits = field(default_factory=Limits)
    """User index 6."""

    user7: Limits = field(default_factory=Limits)
    """User index 7."""

    @classmethod
    def from_ismrmrd_header(
        cls,
        header: ismrmrdHeader,
        encoding_number: int = 0,
    ) -> Self:
        """
        Extract EncodingLimits from ismrmrd.ismrmrdHeader.

        Parameters
        ----------
        header
            ISMRMRD header
        encoding_number
            Encoding number. An ValueError is raised if the encoding number is out of range for the header.

        Returns
        -------
        Extracted EncodingLimits if header.encoding.encodingLimits is not None,
        otherwise an empty EncodingLimits.
        """
        if not 0 <= encoding_number < len(header.encoding):
            raise ValueError(f'encoding_number must be between 0 and {len(header.encoding)}')
        enc: encodingType = header.encoding[encoding_number]
        if enc.encodingLimits is None:
            return cls()
        return cls.from_ismrmrd_encoding_limits_type(enc.encodingLimits)

    @classmethod
    def from_ismrmrd_encoding_limits_type(cls, encoding_limits: encodingLimitsType) -> Self:
        """Generate EncodingLimits from ismrmrd.encodingLimitsType."""
        values = {
            field.name: Limits.from_ismrmrd(getattr(encoding_limits, field.name))
            for field in dataclasses.fields(encoding_limits)
        }

        # adjust from ISMRMRD to MRPro naming convention
        values['k0'] = values.pop('kspace_encoding_step_0')
        values['k1'] = values.pop('kspace_encoding_step_1')
        values['k2'] = values.pop('kspace_encoding_step_2')
        values['user0'] = values.pop('user_0')
        values['user1'] = values.pop('user_1')
        values['user2'] = values.pop('user_2')
        values['user3'] = values.pop('user_3')
        values['user4'] = values.pop('user_4')
        values['user5'] = values.pop('user_5')
        values['user6'] = values.pop('user_6')
        values['user7'] = values.pop('user_7')

        return cls(**values)

    def to_ismrmrd_encoding_limits_type(self) -> encodingLimitsType:
        """Convert EncodingLimits to encodingLimitsType."""
        values = {field.name: Limits.to_ismrmrd(getattr(self, field.name)) for field in dataclasses.fields(self)}

        # k0 is not supported
        values.pop('k0')

        # adjust from MRPro to ISMRMRD naming convention
        values['kspace_encoding_step_1'] = values.pop('k1')
        values['kspace_encoding_step_2'] = values.pop('k2')
        values['user_0'] = values.pop('user0')
        values['user_1'] = values.pop('user1')
        values['user_2'] = values.pop('user2')
        values['user_3'] = values.pop('user3')
        values['user_4'] = values.pop('user4')
        values['user_5'] = values.pop('user5')
        values['user_6'] = values.pop('user6')
        values['user_7'] = values.pop('user7')

        return encodingLimitsType(**values)



"""Zero functional."""

from collections.abc import Sequence

import torch

from mrpro.operators.Functional import ElementaryProximableFunctional, throw_if_negative_or_complex


class ZeroFunctional(ElementaryProximableFunctional):
    """The constant zero functional."""

    def __call__(self, x: torch.Tensor) -> tuple[torch.Tensor,]:
        """Evaluate the zero functional for the given input tensor.

        This functional always returns a tensor of zeros.


        Parameters
        ----------
        x
            Input tensor.

        Returns
        -------
            Tensor of zeros. The shape is determined by the input `x` and the `dim` and `keepdim` at initialization.
            If `dim` is `None`, the shape matches `x`. Else dimensions of `x` indexed by `dim` are reduced to 1 if
            `keepdim` is `True`, otherwise they are removed.

        """
        return super().__call__(x)

    def forward(self, x: torch.Tensor) -> tuple[torch.Tensor,]:
        """Apply forward of ZeroFunctional.

        .. note::
            Prefer calling the instance of the ZeroFunctional as ``operator(x)`` over directly calling this method.
            See this PyTorch `discussion <https://discuss.pytorch.org/t/is-model-forward-x-the-same-as-model-call-x/33460/3>`_.
        """
        # To ensure that the dtype matches what it would be if we were to apply the weight and target
        dtype = torch.promote_types(torch.promote_types(x.dtype, self.weight.dtype), self.target.dtype).to_real()

        if self.dim is None:
            normal_dim: Sequence[int] = range(x.ndim)
        elif not all(-x.ndim <= d < x.ndim for d in self.dim):
            raise IndexError('Invalid dimension index')
        else:
            normal_dim = [d % x.ndim for d in self.dim] if x.ndim > 0 else []

        if self.keepdim:
            new_shape = [1 if i in normal_dim else s for i, s in enumerate(x.shape)]
        else:
            new_shape = [s for i, s in enumerate(x.shape) if i not in normal_dim]

        return (torch.zeros(new_shape, dtype=dtype, device=self.target.device),)

    def prox(self, x: torch.Tensor, sigma: float | torch.Tensor = 1.0) -> tuple[torch.Tensor,]:
        """Apply the proximal operator to a tensor.

        Always returns x, as the proximal operator of a constant functional is the identity.

        Parameters
        ----------
        x
            input tensor
        sigma
            step size

        Returns
        -------
            Result of the proximal operator applied to x
        """
        throw_if_negative_or_complex(sigma)
        dtype = torch.promote_types(torch.promote_types(x.dtype, self.weight.dtype), self.target.dtype)
        return (x.to(dtype=dtype),)

    def prox_convex_conj(self, x: torch.Tensor, sigma: float | torch.Tensor = 1.0) -> tuple[torch.Tensor,]:
        r"""Apply the proximal operator of the convex conjugate of the functional to a tensor.

        The convex conjugate of the zero functional is the indicator function over :math:`C^N \setminus \{0\}`,
        which evaluates to infinity for all values of `x` except zero.
        If ``sigma > 0``, the proximal operator of the scaled convex conjugate is constant zero,
        otherwise it is the identity.

        Parameters
        ----------
        x
            input tensor
        sigma
            step size

        Returns
        -------
            Result of the proximal operator of the convex conjugate applied to x
        """
        throw_if_negative_or_complex(sigma)
        sigma = torch.as_tensor(sigma)
        dtype = torch.promote_types(torch.promote_types(x.dtype, self.weight.dtype), self.target.dtype)
        result = torch.where(sigma == 0, x, torch.zeros_like(x)).to(dtype=dtype)
        return (result,)



"""Summarize the values of a tensor to a string."""

from collections.abc import Sequence
from typing import Protocol, runtime_checkable

import torch


@runtime_checkable
class HasShortStr(Protocol):
    """Object implemtining shortstr."""

    def __shortstr__(self) -> str:
        """Return a short string representation."""


def summarize_object(obj: object) -> str:
    """Summarize object to a human readable representation.

    Parameters
    ----------
    obj
        The object to summarize.
        This method has special cases:
            For sequences of numeric values and tensors, a summary of shape and value is returned.
            If __shortstr__ is implemented, it is used.
            For torch.nn.Modules, only the name is used.
            For other objects, obj.__str__() will be used.

    Returns
    -------
        The string summary/short representation.

    """
    if isinstance(obj, torch.Tensor):
        return summarize_values(obj)
    if isinstance(obj, tuple | list):
        try:
            torch.as_tensor(obj)
        except Exception:  # noqa: BLE001
            return str(obj)
        return summarize_values(obj)
    if isinstance(obj, HasShortStr):
        return obj.__shortstr__()
    if isinstance(obj, torch.nn.Module):
        return f'{type(obj).__name__}(...)'
    return str(obj)


def summarize_values(value: torch.Tensor | Sequence[float], summarization_threshold: int = 6) -> str:
    """Summarize the values of a tensor or sequence to a string.

    Returns a string representation of the object and its values.

    Parameters
    ----------
    value
        The object to summarize.
    summarization_threshold
        The threshold of total array elements triggering summarization.

    Returns
    -------
        A string summary of the shape and values.
    """
    string = []
    if isinstance(value, torch.Tensor):
        if value.shape:
            string.append(f'Tensor<{", ".join(map(str, value.shape))}>,')
        elif value.numel():
            string.append('Tensor:')
        else:
            return 'Tensor'
    elif not len(value):
        return f'{type(value).__name__}'
    else:
        string.append(f'{type(value).__name__}<{len(value)}>,')
    value = torch.as_tensor(value, device='cpu')
    constant = False
    if value.numel() == 0:
        pass
    elif value.is_complex():
        if value.isclose(value.ravel()[0]).all():
            string.append(f'constant {value.ravel()[0].item():.3g}')
            constant = True
        elif value.numel() > summarization_threshold:
            magnitude = value.abs()
            string.append(
                f'|x| ∈ [{magnitude.amin().item():.3g}, {magnitude.amax().item():.3g}], μ={value.mean().item():.3g},'
            )
    else:  # real valued
        min_value, max_value = value.min(), value.max()
        if torch.isclose(min_value, max_value):
            if min_value.dtype == torch.bool:
                string.append(f'constant {min_value.item()}')
            else:
                string.append(f'constant {min_value.item():.3g}')
            constant = True
        elif value.numel() > summarization_threshold:
            mean = (value + 0.0).mean()  # +0. to force float for dtype.int64
            string.append(f'x ∈ [{min_value.item():.3g}, {max_value.item():.3g}], μ={mean.item():.3g},')
    if not constant:  # no need to add values for constant tensor
        edgeitems = 1 if summarization_threshold < 4 else 2
        with torch._tensor_str.printoptions(threshold=summarization_threshold, edgeitems=edgeitems):
            values = ''.join(
                torch._tensor_str._tensor_str(torch.as_tensor(value.ravel(), device='cpu'), 0).splitlines()
            )
            string.append(values)
    return ' '.join(string)



"""Returns the trajectory saved in an ISMRMRD raw data file."""

from collections.abc import Sequence

import ismrmrd
import torch

from mrpro.data.KTrajectory import KTrajectory
from mrpro.data.SpatialDimension import SpatialDimension
from mrpro.utils.reshape import unsqueeze_at


class KTrajectoryIsmrmrd:
    """Get trajectory in ISMRMRD raw data file.

    Use an instance of this class to tell `mrpro.data.KData.from_file` to read in the trajectory
    from the ISMRMRD file [TRA]_.

    The trajectory is optionally normalized to fit in the encoding matrix.

    References
    ----------
    .. [TRA] ISMRMRD trajectory https://ismrmrd.readthedocs.io/en/latest/mrd_raw_data.html#k-space-trajectory
    """

    def __init__(self, normalize: bool = False) -> None:
        """Initialize KTrajectoryIsmrmrd.

        The ISMRMRD specification deos not define a normalization for the trajectory. If the trajectory is not
        normalized to the encoding_matrix, please set ``normalize=True`` `to enable normalization. Otherwise,
        the trajectory is read in as-is.

        Parameters
        ----------
        normalize
            normalize the trajectory to the encoding matrix.
        """
        self.normalize = normalize

    def __call__(self, acquisitions: Sequence[ismrmrd.Acquisition], encoding_matrix: SpatialDimension) -> KTrajectory:
        """Read out the trajectory from the ISMRMRD data file.

        Parameters
        ----------
        acquisitions
            list of ismrmrd acquisistions to read from. Needs at least one acquisition.
        encoding_matrix
            encoding matrix, used to normalize the trajectory.

        Returns
        -------
            trajectory in the shape of the original raw data.
        """
        traj = torch.stack([torch.as_tensor(acq.traj, dtype=torch.float32) for acq in acquisitions])

        if not traj.numel():
            raise ValueError('No trajectory information available in the ISMRMD file.')

        if traj.shape[-1] != 3:  # ISMRMRD can contain 2D trajectories. We enforce 3D
            zero = torch.zeros_like(traj[..., :1])
            traj = torch.cat([traj, *([zero] * (3 - traj.shape[-1]))], dim=-1)

        traj = unsqueeze_at(traj, dim=-3, n=5 - traj.ndim + 1)  # +1 due to stack dim
        scaling = encoding_matrix if self.normalize else None
        return KTrajectory.from_tensor(traj, stack_dim=-1, axes_order='xyz', scaling_matrix=scaling)



"""Radial phase encoding (RPE) trajectory class."""

import torch

from mrpro.data.KTrajectory import KTrajectory
from mrpro.data.traj_calculators.KTrajectoryCalculator import KTrajectoryCalculator
from mrpro.utils.reshape import unsqueeze_tensors_left


class KTrajectoryRpe(KTrajectoryCalculator):
    """Radial phase encoding trajectory.

    Frequency encoding along kx is carried out in a standard Cartesian way. The phase encoding points along ky and kz
    are positioned along radial lines [BOU2009]_ [KOL2014]_.

    References
    ----------
    .. [BOU2009] Boubertakh R, Schaeffter T (2009) Whole-heart imaging using undersampled radial phase encoding (RPE)
        and iterative sensitivity encoding (SENSE) reconstruction. MRM 62(5) https://doi.org/10.1002/mrm.22102

    .. [KOL2014] Kolbitsch C, Schaeffter T (2014) A 3D MR-acquisition scheme for nonrigid bulk motion correction
        in simultaneous PET-MR. Medical Physics 41(8) https://doi.org/10.1118/1.4890095
    """

    def __init__(self, angle: float, shift_between_rpe_lines: tuple | torch.Tensor = (0, 0.5, 0.25, 0.75)) -> None:
        """Initialize KTrajectoryRpe.

        Parameters
        ----------
        angle
            angle in rad between two radial phase encoding lines
        shift_between_rpe_lines
            shift between radial phase encoding lines along the radial direction.
            See _apply_shifts_between_rpe_lines() for more details
        """
        super().__init__()

        self.angle: float = angle
        self.shift_between_rpe_lines: torch.Tensor = torch.as_tensor(shift_between_rpe_lines)

    def _apply_shifts_between_rpe_lines(self, radial: torch.Tensor, idx: torch.Tensor) -> torch.Tensor:
        """Shift radial phase encoding lines relative to each other.

        Example: shift_between_rpe_lines = [0, 0.5, 0.25, 0.75] leads to a shift of the 0th line by 0,
        the 1st line by 0.5, the 2nd line by 0.25, the 3rd line by 0.75, the 4th line by 0, the 5th line
        by 0.5 and so on. Phase encoding points in k-space center are not shifted [PRI2010]_.

        Line #          k-space points before shift             k-space points after shift
        0               +    +    +    +    +    +    +         +    +    +    +    +    +    +
        1               +    +    +    +    +    +    +           +    +    +  +      +    +    +
        2               +    +    +    +    +    +    +          +    +    +   +     +    +    +
        3               +    +    +    +    +    +    +            +    +    + +       +    +    +
        4               +    +    +    +    +    +    +         +    +    +    +    +    +    +
        5               +    +    +    +    +    +    +           +    +    +  +      +    +    +

        Parameters
        ----------
        radial
            k-space positions along each phase encoding line, to be shifted
        idx
            indices used for shift calculation

        Returns
        -------
            shifted radial k-space positions

        References
        ----------
        .. [PRI2010] Prieto C, Schaeffter T (2010) 3D undersampled golden-radial phase encoding
        for DCE-MRA using inherently regularized iterative SENSE. MRM 64(2). https://doi.org/10.1002/mrm.22446
        """
        radial, idx = torch.broadcast_tensors(radial, idx)
        radial = radial.clone()
        not_center = ~torch.isclose(radial, torch.tensor(0.0))
        for ind, shift in enumerate(self.shift_between_rpe_lines):
            current_mask = (idx % len(self.shift_between_rpe_lines)) == ind
            # k-space center should not be shifted
            current_mask &= not_center
            radial[current_mask] += shift

        return radial

    def __call__(
        self,
        *,
        n_k0: int,
        k0_center: int | torch.Tensor,
        k1_idx: torch.Tensor,
        k1_center: int | torch.Tensor,
        k2_idx: torch.Tensor,
        reversed_readout_mask: torch.Tensor | None = None,
        **_,
    ) -> KTrajectory:
        """Calculate radial phase encoding trajectory for given header information.

        Parameters
        ----------
        n_k0
            number of samples in k0
        k0_center
            position of k-space center in k0
        k1_idx
            indices of k1
        k1_center
            position of k-space center in k1
        k2_idx
            indices of k2
        reversed_readout_mask
            boolean tensor indicating reversed readout

        Returns
        -------
            radial phase encoding trajectory for given header information
        """
        angles = k2_idx * self.angle

        radial = (k1_idx - k1_center).to(torch.float32)
        radial = self._apply_shifts_between_rpe_lines(radial, k2_idx)

        kz = radial * torch.sin(angles)
        ky = radial * torch.cos(angles)
        kx = self._readout(n_k0, k0_center, reversed_readout_mask=reversed_readout_mask)
        kz, ky, kx = unsqueeze_tensors_left(kz, ky, kx, ndim=5)
        return KTrajectory(kz, ky, kx)



"""MR image data (IData) class."""

import datetime
import re
import warnings
from collections.abc import Generator, Sequence
from pathlib import Path

import nibabel
import numpy as np
import pydicom
import torch
from einops import rearrange, repeat
from pydicom import dcmread
from pydicom.dataset import Dataset
from pydicom.pixels import set_pixel_data
from typing_extensions import Self

from mrpro.data.Dataclass import Dataclass
from mrpro.data.IHeader import IHeader, get_items
from mrpro.data.KHeader import KHeader
from mrpro.data.SpatialDimension import SpatialDimension
from mrpro.utils.reshape import unsqueeze_right
from mrpro.utils.summarize import summarize_values
from mrpro.utils.unit_conversion import m_to_mm, rad_to_deg, s_to_ms


def _dcm_pixelarray_to_tensor(dataset: Dataset) -> torch.Tensor:
    """Transform pixel array in dicom file to tensor.

    Rescale intercept, (0028|1052), and rescale slope (0028|1053) are
    DICOM tags that specify the linear transformation from pixels in
    their stored on disk representation to their in memory
    representation. U = m*SV + b where U is in output units, m is
    the rescale slope, SV is the stored value, and b is the rescale
    intercept [RES]_.

    Rescale intercept and rescale slope can be defined for each frame in a 3D/M2D dicom file, as a global value or
    not at all.

    References
    ----------
    .. [RES] Rescale intercept and slope https://www.kitware.com/dicom-rescale-intercept-rescale-slope-and-itk/
    """
    slope = sl if (sl := get_items([dataset], 'RescaleSlope', lambda x: float(x))) else [1.0]
    slope_tensor = unsqueeze_right(torch.as_tensor(slope), dataset.pixel_array.ndim - 1)
    intercept = ic if (ic := get_items([dataset], 'RescaleIntercept', lambda x: float(x))) else [0.0]
    intercept_tensor = unsqueeze_right(torch.as_tensor(intercept), dataset.pixel_array.ndim - 1)

    # Image data is 2D or 3D np.array of Uint16, which cannot directly be converted to tensor
    return slope_tensor * torch.as_tensor(dataset.pixel_array.astype(np.complex64)) + intercept_tensor


def _natural_key(s: str | Path) -> list[int | str]:
    """Key for natural sorting of strings with numbers.

    Example:
        >>> strings = ['img_1.dcm', 'img_10.dcm', 'img_2.dcm']
        >>> sorted(strings, key=_natural_key)
        ['img_1.dcm', 'img_2.dcm', 'img_10.dcm']
    """
    return [int(t) if t.isdigit() else t for t in re.split(r'(\d+)', str(s))]


class IData(Dataclass):
    """MR image data (IData) class."""

    data: torch.Tensor
    """Tensor containing image data with dimensions `(*other, coils, z, y, x)`."""

    header: IHeader
    """Header for image data."""

    def rss(self, keepdim: bool = False) -> torch.Tensor:
        """Root-sum-of-squares over coils image data.

        Parameters
        ----------
        keepdim
            if `True`, the output tensor has the same number of dimensions as the data tensor, and the coil dimension is
            kept as a singleton dimension. If `False`, the coil dimension is removed.

        Returns
        -------
            image data tensor with shape `(..., 1, z, y, x)` if `keepdim` is `True`
            or `(..., z, y, x)` if keepdim is `False`.
        """
        coildim = -4
        return self.data.abs().square().sum(dim=coildim, keepdim=keepdim).sqrt()

    @classmethod
    def from_tensor_and_kheader(cls, data: torch.Tensor, header: KHeader) -> Self:
        """Create IData object from a tensor and a KHeader object.

        Parameters
        ----------
        data
            image data with dimensions (broadcastable to) `(other, coils, z, y, x)`.
        header
            MR raw data header containing required meta data for the image header.
        """
        iheader = IHeader.from_kheader(header)
        return cls(header=iheader, data=data)

    @classmethod
    def from_dicom_files(cls, filenames: Sequence[str | Path] | Generator[Path, None, None] | str | Path) -> Self:
        """Read multiple DICOM files and return IData object.

        DICOM images can be saved as single-frame or multi-frame images [DCMMF]_.

        If the DICOM files are single-frame, we treat each file separately and stack them along the `other` dimension.
        If the DICOM files are multi-frame and the MRAcquisitionType is ``3D`` we treat the frame dimension as the `z`
        dimension. Otherwise, we move the frame dimension to the `other` dimension. Multiple multi-frame DICOM
        images are stacked along an additional `other` dimension before the frame dimension.
        Providing the list of files sorted by filename usually leads to a reasonable sorting of the data.

        Parameters
        ----------
        filenames
            Filename or sequence of DICOM filenames.

        References
        ----------
        .. [DCMMF] https://dicom.nema.org/medical/dicom/2020b/output/chtml/part03/sect_C.7.6.16.html
        """
        if isinstance(filenames, str | Path):
            datasets = [dcmread(filenames)]
        else:
            # Use natsort to ensure correct order of filenames like img_1.dcm, img_2.dcm, ..., img_10.dcm
            datasets = [dcmread(filename) for filename in sorted(filenames, key=_natural_key)]
        if not datasets:  # check datasets (not filenames) to allow for filenames to be a Generator
            raise ValueError('No dicom files specified')

        header = IHeader.from_dicom(*datasets)

        # Ensure that data has the same shape and can be stacked
        if not all(ds.pixel_array.shape == datasets[0].pixel_array.shape for ds in datasets):
            raise ValueError('Only dicom files with data of the same shape can be stacked.')

        data = torch.stack([_dcm_pixelarray_to_tensor(ds) for ds in datasets])

        # NumberofFrames (0028|0008): The total number of frames contained within a Multi-frame Image
        number_of_frames = [item.value for ds in datasets for item in ds.iterall() if item.tag == 0x00280008]

        if len(number_of_frames) > 0 and float(number_of_frames[0]) > 1:  # multi-frame data
            # MRAcquisitionType (0018|0023) is 1D/2D/3D
            mr_acquisition_type = [item.value for item in datasets[0].iterall() if item.tag == 0x00180023]

            if len(mr_acquisition_type) > 0 and mr_acquisition_type[0] == '3D':  # multi-frame 3D data
                data = repeat(data, 'other z x y -> other coils z y x', coils=1)
            else:  # multi-frame 2D data, rearrange data and header
                if data.shape[0] == 1:  # if other dimension is singleton, there is only frame dimension
                    data = repeat(data, '1 frame x y -> frame coils z y x', coils=1, z=1)
                else:
                    data = repeat(data, 'other frame x y -> other frame coils z y x', coils=1, z=1)
                    header = header.rearrange('other coils frame x y -> other frame coils z y x', z=1)
        else:  # single-frame data
            data = repeat(data, 'other x y -> other coils z y x', coils=1, z=1)

        return cls(data=data, header=header)

    @classmethod
    def from_dicom_folder(cls, foldername: str | Path, suffix: str | None = 'dcm') -> Self:
        """Read all DICOM files from a folder and return IData object.

        Parameters
        ----------
        foldername
            path to folder with DICOM files.
        suffix
            file extension (without period/full stop) to identify the DICOM files.
            If `None`, then all files in the folder are read in.
        """
        file_paths = list(Path(foldername).glob('*')) if suffix is None else list(Path(foldername).glob('*.' + suffix))
        if len(file_paths) == 0:
            raise ValueError(f'No dicom files with suffix {suffix} found in {foldername}')

        # Pass on sorted file list as order of dicom files is often the same as the required order
        return cls.from_dicom_files(filenames=sorted(file_paths))

    def to_nifti(self, filename: str | Path, magnitude_only: bool = False) -> None:
        """Save image data as NIFTI file.

        Parameters
        ----------
        filename
            filename / path of the NIFTI file.
        magnitude_only
            if `True`, only the magnitude of the image data is saved as a NIFTI1 file,
            otherwise the complex image data is saved as a NIFTI2 file.
            Many software packages do not support complex data.
        """
        orientation = self.header.orientation.mean().as_matrix()
        position = torch.stack([p.mean() for p in self.header.position.zyx])
        affine_zyx = torch.cat(
            [torch.tensor([[1.0, 0.0, 0.0, 0.0]]), torch.cat([position[:, None], orientation], 1)], 0
        )
        affine = affine_zyx.flip([0, 1])
        data = rearrange(self.data, '... other coils z y x-> x y z 1 other (...) coils')
        if magnitude_only:
            nifti = nibabel.nifti2.Nifti1Image(data.abs().numpy(), affine.numpy(), dtype=np.float32)
        else:
            nifti = nibabel.nifti2.Nifti2Image(data.numpy(), affine.numpy(), dtype=np.complex64)

        nifti.header['pixdim'][1:4] = [self.header.resolution.x, self.header.resolution.y, self.header.resolution.z]
        description = (
            f'TE={summarize_values(self.header.te)}ms; '
            f'TI={summarize_values(self.header.ti)}ms; '
            f'TR={summarize_values(self.header.tr)}ms; '
            f'FA={summarize_values(self.header.fa)}rad'
        )
        nifti.header['descrip'] = description.encode('utf-8')
        nifti.to_filename(filename)

    def __repr__(self):
        """Representation method for IData class."""
        try:
            device = str(self.device)
        except RuntimeError:
            device = 'mixed'
        out = (
            f'{type(self).__name__} with shape: {list(self.data.shape)!s} and dtype {self.data.dtype}\n'
            f'Device: {device}\n{self.header}'
        )
        return out

    def to_dicom_folder(
        self,
        foldername: str | Path,
        *,
        series_description: str | None = None,
        reference_patient_table_position: SpatialDimension | None = None,
        rescale_slope: float = 1.0,
        rescale_intercept: float = 0.0,
        normalize_data: bool = True,
    ) -> None:
        """Write the image data to DICOM files in a folder.

        The data is always saved in a multi-frame DICOM files.

        Parameters
        ----------
        foldername
            Path to folder for DICOM files.
        series_description
            String to be saved as the series description in the DICOM files.
        reference_patient_table_position
            If provided, the image position is calculated relative to this table positiion. This ensures that the
            image position is consistent across different scans even if the patient table has moved.
        rescale_slope
            Slope of linear scaling of data. Data is save as (data - intercept)/slope.
        rescale_intercept
            Intercept of linear scaling of data. Data is save as (data - intercept)/slope.
        normalize_data
            Normalize data prior to applying scaling.
        """
        if not isinstance(foldername, Path):
            foldername = Path(foldername)
        foldername.mkdir(parents=True, exist_ok=False)

        # We save 3D image data in each dicom file. This can either be a full 3D volume, multiple slices (M2D) or
        # a combination of y and x image dimensions an one other dimension, e.g. multiple cardiac phases of a 2D image.
        mr_acquisition_type = '3D' if self.data.shape[-3] > 1 else '2D'
        frame_dimension = next((i for i in range(-3, -len(self.data.shape) - 1, -1) if self.data.shape[i] > 1), -3)
        number_of_frames = self.data.shape[frame_dimension]
        dcm_idata = self.swapdims(frame_dimension, -3)

        # Metadata
        file_meta = pydicom.dataset.FileMetaDataset()
        file_meta.MediaStorageSOPClassUID = pydicom.uid.MRImageStorage
        file_meta.MediaStorageSOPInstanceUID = pydicom.uid.generate_uid()
        file_meta.TransferSyntaxUID = pydicom.uid.ExplicitVRLittleEndian

        # Dataset
        dataset = pydicom.Dataset()
        dataset.file_meta = file_meta

        dataset.PatientName = 'Unknown'
        dataset.PatientID = 'Unknown'
        dataset.PatientSex = 'O'
        dataset.Modality = 'MR'
        dataset.StudyDescription = 'MRpro'

        timestamp = self.header.datetime or datetime.datetime.now(datetime.timezone.utc)
        dataset.SeriesDate = timestamp.strftime('%Y%m%d')
        dataset.SeriesTime = timestamp.strftime('%H%M%S.%f')
        if series_description:
            dataset.SeriesDescription = series_description
            dataset.ProtocolName = series_description
        dataset.SeriesInstanceUID = pydicom.uid.generate_uid()

        dataset.PatientPosition = 'HFS'

        for file_index, other in enumerate(np.ndindex(dcm_idata.shape[:-3])):
            dcm_file_idata = dcm_idata[(*other, slice(None), slice(None), slice(None))]

            dataset.MRAcquisitionType = mr_acquisition_type
            dataset.PerFrameFunctionalGroupsSequence = pydicom.Sequence()

            for frame in range(number_of_frames):
                dcm_frame_idata = dcm_file_idata[..., frame, :, :]
                frame_info = Dataset()
                if dcm_frame_idata.header.shape.numel() != 1:
                    raise ValueError('Only single image can be saved as a frame.')
                directions = dcm_frame_idata.header.orientation[0].as_directions()
                readout_direction = torch.stack(directions[2].zyx[::-1]).squeeze().numpy()
                phase_direction = torch.stack(directions[1].zyx[::-1]).squeeze().numpy()
                slice_direction = torch.stack(directions[0].zyx[::-1]).squeeze().numpy()
                position = torch.stack(dcm_frame_idata.header.position.zyx[::-1]).squeeze().numpy()

                image_position_patient = (
                    position
                    - readout_direction * dcm_frame_idata.header.resolution.x * dcm_frame_idata.data.shape[-1] / 2
                    - phase_direction * dcm_frame_idata.header.resolution.y * dcm_frame_idata.data.shape[-2] / 2
                )
                if mr_acquisition_type == '3D':
                    image_position_patient = (
                        image_position_patient
                        + slice_direction * dcm_frame_idata.header.resolution.z * (-number_of_frames / 2 + frame)
                    )
                if reference_patient_table_position:
                    image_position_patient += np.squeeze(
                        torch.stack(
                            (dcm_frame_idata.header.patient_table_position - reference_patient_table_position).zyx[::-1]
                        ).numpy()
                    )
                position_ds = Dataset()
                position_ds.ImagePositionPatient = m_to_mm(image_position_patient.tolist())
                frame_info.PlanePositionSequence = pydicom.Sequence([position_ds])

                # According to the dicom manual, ImageOrientationPatient describes:
                # "The direction cosines of the first row and the first column with respect to the patient."
                # This would suggest that the first direction should be the readout direction, the second the
                # phase direction. Nevertheless, for all our data the two directions have to be swapped to achieve
                # the correct orientation. Any help welcome in solving this!
                orientation_ds = Dataset()
                orientation_ds.ImageOrientationPatient = [*phase_direction, *readout_direction]
                frame_info.PlaneOrientationSequence = pydicom.Sequence([orientation_ds])

                def get_singleton(parameter: torch.Tensor | list[float], parameter_name: str) -> float | None:
                    """Return unique value of parameter tensor or list. Raise warning if not unique."""
                    unique_values = torch.unique(torch.as_tensor(parameter))
                    if unique_values.numel() > 1:
                        warnings.warn(
                            f'{parameter_name} is not singleton. Using first value. To ensure all values are saved, '
                            'split data into correct subsets first.',
                            stacklevel=2,
                        )
                    return unique_values[0].item() if len(unique_values) > 0 else None

                echo_ds = Dataset()
                if (echo_time := get_singleton(dcm_frame_idata.header.te, 'te')) is not None:
                    echo_ds.EchoTime = s_to_ms(echo_time)
                frame_info.MREchoSequence = pydicom.Sequence([echo_ds])

                pixel_measures = Dataset()
                resolution = dcm_frame_idata.header.resolution[0]
                pixel_measures.SliceThickness = m_to_mm(resolution.z)
                pixel_measures.PixelSpacing = m_to_mm([resolution.x, resolution.y])
                frame_info.PixelMeasuresSequence = pydicom.Sequence([Dataset(pixel_measures)])

                timing_parameters = Dataset()
                if (flip_angle := get_singleton(dcm_frame_idata.header.fa, 'fa')) is not None:
                    timing_parameters.FlipAngle = rad_to_deg(flip_angle)
                if (inversion_time := get_singleton(dcm_frame_idata.header.ti, 'ti')) is not None:
                    timing_parameters.InversionTime = s_to_ms(inversion_time)
                if (repetition_time := get_singleton(dcm_frame_idata.header.tr, 'tr')) is not None:
                    timing_parameters.RepetitionTime = s_to_ms(repetition_time)
                frame_info.MRTimingAndRelatedParametersSequence = pydicom.Sequence([timing_parameters])

                pixel_value_transformation = Dataset()
                pixel_value_transformation.RescaleSlope = rescale_slope
                pixel_value_transformation.RescaleIntercept = rescale_intercept
                pixel_value_transformation.RescaleType = 'US'
                frame_info.PixelValueTransformationSequence = pydicom.Sequence([pixel_value_transformation])

                dataset.PerFrameFunctionalGroupsSequence.append(frame_info)

            # (frames, rows, columns) for multi-frame grayscale data
            pixel_data = dcm_file_idata.data.abs().cpu().numpy()
            pixel_data = pixel_data / pixel_data.max() * (2**16 - 1) if normalize_data else pixel_data
            pixel_data = (pixel_data - rescale_intercept) / rescale_slope
            pixel_data = rearrange(pixel_data[0, 0, ...], 'frames y x -> frames x y')

            if np.any(clipped_idx := (pixel_data < 0) | (pixel_data > 65535)):
                clipped_data = pixel_data[clipped_idx]
                warnings.warn(
                    'Values outside of the uint16 range will be clipped. '
                    + f'Data range [{clipped_data.min()} - {clipped_data.max()}]',
                    stacklevel=2,
                )

            # 'MONOCHROME2' means smallest value is black, largest value is white
            set_pixel_data(
                ds=dataset, arr=pixel_data.astype(np.uint16), photometric_interpretation='MONOCHROME2', bits_stored=16
            )

            dataset.save_as(foldername / f'im_mrpro_{np.prod(file_index)}.dcm', enforce_file_format=True)



# This file is used to get the version of the package from the VERSION file
from importlib.metadata import PackageNotFoundError, version
from pathlib import Path

try:
    __version__ = version('mrpro')
except PackageNotFoundError:
    __version__ = Path(__file__).parent.joinpath('VERSION').read_text().strip()
